# Copyright 2009 Brian Quinlan. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""Execute computations asynchronously using threads or processes."""
from __future__ import absolute_import

__author__ = 'Brian Quinlan (brian@sweetapp.com)'

from ._base import (FIRST_COMPLETED,
                    FIRST_EXCEPTION,
                    ALL_COMPLETED,
                    CancelledError,
                    TimeoutError,
                    Future,
                    Executor,
                    wait,
                    as_completed)
from .thread import ThreadPoolExecutor

# Jython doesn't have multiprocessing
try:
    from .process import ProcessPoolExecutor
except ImportError:
    pass



from enstaller.repository import Repository, RepositoryPackageMetadata
from enstaller.utils import PY_VER


def parse_index(json_dict, store_location, python_version=PY_VER):
    """
    Parse the given json index data and iterate package instance over its
    content.

    Parameters
    ----------
    json_dict: dict
        Parsed legacy json index
    store_location: str
        A label describing where the index is coming from
    python_version: str
        The major.minor string describing the python version. This generator
        will iterate over every package where the python attribute is `null` or
        equal to this string. If python_version == "*", then every package is
        iterated over.
    """
    for key, info in json_dict.items():
        info.setdefault('type', 'egg')
        info.setdefault('packages', [])
        info["store_location"] = store_location
        info.setdefault('python', python_version)
        if python_version == "*":
            yield RepositoryPackageMetadata.from_json_dict(key, info)
        elif info["python"] in (None, python_version):
            yield RepositoryPackageMetadata.from_json_dict(key, info)


def repository_factory(session, indices):
    """
    Create a repository from legacy indices.

    Parameters
    ----------
    session : Session
        A :py:class:`~enstaller.session.Session` instance.
    indices : list
        Sequence of (index_url, store_name) pairs, e.g. the indices property of
        :py:class:`~enstaller.config.Configuration` instances.

    Note
    ----

    This does not use etag caching nor write anything in sys.stdout. If you
    want to use etag caching, simply do::

        with session.etag():
            repository = repository_factory(session, ...)
    """
    repository = Repository()
    for url, store_location in indices:
        resp = session.fetch(url)
        json_data = resp.json()

        for package in parse_index(json_data, store_location):
            repository.add_package(package)
    return repository



import os
import functools

class Preference:
	"""
	A decorator wrapping a 'priority' classmethod
	"""
	def __init__(self, name):
		"""
		Create a decorator giving preference to XDG_CURRENT_DESKTOP of 'name'
		"""
		self.name = name

	def decorate(self, func):
		self.func = func
		@functools.wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, **kwargs) * self.multiplier
		return wrapper
	__call__ = decorate

	@property
	def multiplier(self):
		matches = os.environ.get('XDG_CURRENT_DESKTOP') == self.name
		return 1.5 if matches else 1



from __future__ import print_function

import collections
import contextlib
import json
import sys

import mock

from egginst.vendor.six import PY2, StringIO

from enstaller.config import Configuration
from enstaller.enpkg import Enpkg
from enstaller.plat import custom_plat
from enstaller.repository import (InstalledPackageMetadata, Repository,
                                  RepositoryPackageMetadata)
from enstaller.session import Session
from enstaller.utils import PY_VER
from enstaller.vendor import responses
from enstaller.versions.enpkg import EnpkgVersion


FAKE_MD5 = "a" * 32
FAKE_SIZE = -1


R_JSON_AUTH_RESP = {
    'first_name': u'David',
    'has_subscription': True,
    'is_active': True,
    'is_authenticated': True,
    'last_name': u'Cournapeau',
    'subscription_level': u'basic'
}

R_JSON_AUTH_FREE_RESP = {
    'first_name': u'David',
    'has_subscription': False,
    'is_active': True,
    'is_authenticated': True,
    'last_name': u'Cournapeau',
    'subscription_level': u'free'
}

R_JSON_NOAUTH_RESP = {
    'is_authenticated': False,
    'last_name': u'Cournapeau',
    'first_name': u'David',
    'has_subscription': True,
    'subscription_level': u'basic'
}

SIMPLE_INDEX = {
    "nose-1.3.4-1.egg": {
        "available": True,
        "build": 1,
        "md5": "7fbd5a7c83ebbb14c42141ed734505ef",
        "mtime": 1409944509.0,
        "name": "nose",
        "product": "free",
        "python": None,
        "size": 334992,
        "type": "egg",
        "version": "1.3.4"
    },
}


if sys.version_info < (2, 7):
    # FIXME: this looks quite fishy. On 2.6, with unittest2, the assertRaises
    # context manager does not contain the actual exception object ?
    def exception_code(ctx):
        return ctx.exception
else:
    def exception_code(ctx):
        return ctx.exception.code


class DummyAuthenticator(object):
    def __init__(self, user_info=None):
        self._auth = None

    @property
    def auth(self):
        return self._auth

    def authenticate(self, session, auth):
        self._auth = auth


def dummy_installed_package_factory(name, version, build, key=None,
                                    py_ver=PY_VER, store_location=""):
    key = key if key else "{0}-{1}-{2}.egg".format(name, version, build)
    version = EnpkgVersion.from_upstream_and_build(version, build)
    return InstalledPackageMetadata(key, name.lower(), version, [], py_ver,
                                    "", store_location)


def dummy_repository_package_factory(name, version, build, key=None,
                                     py_ver=PY_VER, store_location="",
                                     dependencies=None, mtime=0.0):
    dependencies = dependencies or []
    key = key if key else "{0}-{1}-{2}.egg".format(name, version, build)
    fake_size = FAKE_SIZE
    fake_md5 = FAKE_MD5
    fake_mtime = mtime
    version = EnpkgVersion.from_upstream_and_build(version, build)
    return RepositoryPackageMetadata(key, name.lower(), version,
                                     dependencies, py_ver, fake_size,
                                     fake_md5, fake_mtime, "commercial",
                                     True, store_location)


def repository_factory(entries):
    repository = Repository()
    for entry in entries:
        repository.add_package(entry)
    return repository


class MockedPrint(object):
    def __init__(self):
        self.s = StringIO()

    def __call__(self, *a):
        self.s.write(" ".join(str(_) for _ in a) + "\n")

    @property
    def value(self):
        return self.s.getvalue()


@contextlib.contextmanager
def mock_print():
    m = MockedPrint()

    if PY2:
        with mock.patch("__builtin__.print", m):
            yield m
    else:
        with mock.patch("builtins.print", m):
            yield m


@contextlib.contextmanager
def mock_input(input_string):
    def f(ignored):
        return input_string

    if PY2:
        with mock.patch("__builtin__.raw_input", f):
            yield f
    else:
        with mock.patch("builtins.input", f):
            yield f


# Decorators to force a certain configuration
def make_keyring_unavailable(f):
    return mock.patch("enstaller.config.keyring", None)(f)


# Context managers to force certain configuration
@contextlib.contextmanager
def make_keyring_unavailable_context():
    with mock.patch("enstaller.config.keyring", None) as context:
        yield context


# Context managers to force certain configuration
@contextlib.contextmanager
def make_keyring_available_context():
    m = mock.Mock(["get_password", "set_password"])
    with mock.patch("enstaller.config.keyring", m) as context:
        yield context


@contextlib.contextmanager
def make_default_configuration_path(path):
    with mock.patch("enstaller.main.get_config_filename",
                    lambda ignored: path) as context:
        yield context


@contextlib.contextmanager
def mock_input_auth(username, password):
    with mock.patch("enstaller.main.input_auth",
                    return_value=(username, password)) as context:
        yield context


class _FakeKeyring(object):
    def __init__(self):
        self._state = collections.defaultdict(dict)

    def get_password(self, service, username):
        if service in self._state:
            return self._state[service].get(username, None)
        else:
            return None

    def set_password(self, service, username, password):
        self._state[service][username] = password


@contextlib.contextmanager
def fake_keyring_context():
    keyring = _FakeKeyring()
    with mock.patch("enstaller.config.keyring.get_password",
                    keyring.get_password):
        with mock.patch("enstaller.config.keyring.set_password",
                        keyring.set_password):
            yield keyring


def fake_keyring(f):
    keyring = _FakeKeyring()
    dec1 = mock.patch("enstaller.config.keyring.get_password",
                      keyring.get_password)
    dec2 = mock.patch("enstaller.config.keyring.set_password",
                      keyring.set_password)
    return dec1(dec2(f))


@contextlib.contextmanager
def mock_history_get_state_context(state=None):
    if state is None:
        state = set()
    with mock.patch("enstaller.enpkg.History") as context:
        context.return_value.get_state.return_value = set(state)
        yield context


@contextlib.contextmanager
def mock_raw_input(return_value):
    def _function(message):
        print(message)
        return return_value

    with mock.patch("enstaller.utils.input",
                    side_effect=_function) as context:
        yield context


def unconnected_enpkg_factory(prefixes=None):
    """
    Create an Enpkg instance which does not require an authenticated
    repository.
    """
    if prefixes is None:
        prefixes = [sys.prefix]
    config = Configuration()
    repository = Repository()
    return Enpkg(repository, mocked_session_factory(config.repository_cache),
                 prefixes=prefixes)


def mocked_session_factory(repository_cache):
    return Session(DummyAuthenticator(), repository_cache)


def create_repositories(remote_entries=None, installed_entries=None):
    if remote_entries is None:
        remote_entries = []
    if installed_entries is None:
        installed_entries = []

    remote_repository = Repository()
    for remote_entry in remote_entries:
        remote_repository.add_package(remote_entry)
    installed_repository = Repository()
    for installed_entry in installed_entries:
        installed_repository.add_package(installed_entry)

    return remote_repository, installed_repository


class FakeOptions(object):
    def __init__(self):
        self.force = False
        self.forceall = False
        self.no_deps = False
        self.yes = False


def create_prefix_with_eggs(config, prefix, installed_entries=None,
                            remote_entries=None):
    if remote_entries is None:
        remote_entries = []
    if installed_entries is None:
        installed_entries = []

    repository = repository_factory(remote_entries)

    enpkg = Enpkg(repository, mocked_session_factory(config.repository_cache),
                  prefixes=[prefix])
    for package in installed_entries:
        package.store_location = prefix
        enpkg._top_installed_repository.add_package(package)
        enpkg._installed_repository.add_package(package)
    return enpkg


def mock_index(index_data, store_url="https://api.enthought.com"):
    def decorator(f):
        @responses.activate
        def wrapped(*a, **kw):
            responses.add(responses.GET,
                          "{0}/accounts/user/info/".format(store_url),
                          body=json.dumps(R_JSON_AUTH_RESP))
            url = "{0}/eggs/{1}/index.json"
            responses.add(responses.GET,
                          url.format(store_url, custom_plat),
                          body=json.dumps(index_data))
            return f(*a, **kw)
        return wrapped
    return decorator


@fake_keyring
def authenticated_config(f):
    config = Configuration()
    config.update(auth=("dummy", "dummy"))

    m = mock.Mock()
    m.return_value = config
    m.from_file.return_value = config

    wrapper = mock.patch("enstaller.main.Configuration", m)
    mock_authenticated_config = mock.patch(
        "enstaller.main.ensure_authenticated_config",
        mock.Mock())
    return mock_authenticated_config(wrapper(f))



from egginst.vendor.six.moves import unittest

from enstaller.repository import PackageVersionInfo
from enstaller.versions.enpkg import EnpkgVersion

from ..requirement import Requirement


V = EnpkgVersion.from_string


class TestRequirement(unittest.TestCase):
    def assertEqualRequirements(self, left, right):
        self.assertEqual(left.as_dict(), right.as_dict())

    def test_init(self):
        for req_string, name, version, build, strictness in [
                ('',          None,  None,  None, 0),
                (' \t',       None,  None,  None, 0),
                ('foo',       'foo', None,  None, 1),
                (u'bar 1.9',  'bar', '1.9', None, 2),
                ('BAZ 1.8-2', 'baz', '1.8', 2,    3),
                ('qux 1.3-0', 'qux', '1.3', 0,    3),
        ]:
            r = Requirement(req_string)
            self.assertEqual(r.name, name)
            self.assertEqual(r.version, version)
            self.assertEqual(r.build, build)
            self.assertEqual(r.strictness, strictness)

    def test_as_dict(self):
        for req_string, d in [
            ('',          dict()),
            ('foo',       dict(name='foo')),
            ('bar 1.9',   dict(name='bar', version='1.9')),
            ('BAZ 1.8-2', dict(name='baz', version='1.8', build=2)),
        ]:
            r = Requirement(req_string)
            self.assertEqual(r.as_dict(), d)

    def test_misc_methods(self):
        for req_string in ['', 'foo', 'bar 1.2', 'baz 2.6.7-5']:
            r = Requirement(req_string)
            self.assertEqual(str(r), req_string)
            self.assertEqual(r, r)
            self.assertEqual(eval(repr(r)), r)

        self.assertNotEqual(Requirement('foo'), Requirement('bar'))
        self.assertNotEqual(Requirement('foo 1.4'), Requirement('foo 1.4-5'))

    def test_matches(self):
        spec = PackageVersionInfo('foo_bar', V('2.4.1-3'))
        for req_string, m in [
            ('', True),
            ('foo', False),
            ('Foo_BAR', True),
            ('foo_Bar 2.4.1', True),
            ('FOO_Bar 1.8.7', False),
            ('FOO_BAR 2.4.1-3', True),
            ('FOO_Bar 2.4.1-1', False),
        ]:
            self.assertEqual(Requirement(req_string).matches(spec), m, req_string)

    def test_from_anything_name(self):
        # Given
        req_arg = "numpy"

        # When
        req = Requirement.from_anything(req_arg)

        # Then
        self.assertEqualRequirements(req, Requirement(req_arg))

    def test_from_anything_name_and_version(self):
        # Given
        req_arg = "numpy 1.8.0"

        # When
        req = Requirement.from_anything(req_arg)

        # Then
        self.assertEqualRequirements(req, Requirement(req_arg))

    def test_from_anything_name_and_version_and_build(self):
        # Given
        req_arg = "numpy 1.8.0-1"

        # When
        req = Requirement.from_anything(req_arg)

        # Then
        self.assertEqualRequirements(req, Requirement(req_arg))

    def test_from_anything_req(self):
        # Given
        req_arg = Requirement("numpy 1.8.0-1")

        # When
        req = Requirement.from_anything(req_arg)

        # Then
        self.assertEqualRequirements(req, req_arg)



import hashlib
import os

from lockfile import FileLock


def _secure_open_write(filename, fmode):
    # We only want to write to this file, so open it in write only mode
    flags = os.O_WRONLY

    # os.O_CREAT | os.O_EXCL will fail if the file already exists, so we only
    #  will open *new* files.
    # We specify this because we want to ensure that the mode we pass is the
    # mode of the file.
    flags |= os.O_CREAT | os.O_EXCL

    # Do not follow symlinks to prevent someone from making a symlink that
    # we follow and insecurely open a cache file.
    if hasattr(os, "O_NOFOLLOW"):
        flags |= os.O_NOFOLLOW

    # On Windows we'll mark this file as binary
    if hasattr(os, "O_BINARY"):
        flags |= os.O_BINARY

    # Before we open our file, we want to delete any existing file that is
    # there
    try:
        os.remove(filename)
    except (IOError, OSError):
        # The file must not exist already, so we can just skip ahead to opening
        pass

    # Open our file, the use of os.O_CREAT | os.O_EXCL will ensure that if a
    # race condition happens between the os.remove and this line, that an
    # error will be raised. Because we utilize a lockfile this should only
    # happen if someone is attempting to attack us.
    fd = os.open(filename, flags, fmode)
    try:
        return os.fdopen(fd, "wb")
    except:
        # An error occurred wrapping our FD in a file object
        os.close(fd)
        raise


class FileCache(object):
    def __init__(self, directory, forever=False, filemode=0o0600,
                 dirmode=0o0700):
        self.directory = directory
        self.forever = forever
        self.filemode = filemode
        self.dirmode = dirmode

    @staticmethod
    def encode(x):
        return hashlib.sha224(x.encode()).hexdigest()

    def _fn(self, name):
        hashed = self.encode(name)
        parts = list(hashed[:5]) + [hashed]
        return os.path.join(self.directory, *parts)

    def get(self, key):
        name = self._fn(key)
        if not os.path.exists(name):
            return None

        with open(name, 'rb') as fh:
            return fh.read()

    def set(self, key, value):
        name = self._fn(key)

        # Make sure the directory exists
        try:
            os.makedirs(os.path.dirname(name), self.dirmode)
        except (IOError, OSError):
            pass

        with FileLock(name) as lock:
            # Write our actual file
            with _secure_open_write(lock.path, self.filemode) as fh:
                fh.write(value)

    def delete(self, key):
        name = self._fn(key)
        if not self.forever:
            os.remove(name)



"""
A dumb egg unpacker.

It only unpack the egg, and do *not* pre/post process them. You probably do not
want to use this tool if you are not sure about its purpose.
"""
import argparse
import os.path
import posixpath
import sys

from egginst.main import EGG_INFO
from egginst._zipfile import ZipFile


def extract_egg(path, to):
    usr = posixpath.join(EGG_INFO, "usr")

    with ZipFile(path) as zp:
        for p in zp.infolist():
            if p.filename.startswith(usr):
                target = os.path.relpath(p.filename, usr)
                zp.extract_to(p, target, to)


def main(argv=None):
    argv = argv or sys.argv[1:]

    p = argparse.ArgumentParser("A dumb egg unpacker")
    p.add_argument("target",
                   help="Directory where to extract the egg.")
    p.add_argument("eggs", help="Path to the egg.", nargs="+")

    ns = p.parse_args(argv)

    for egg in ns.eggs:
        extract_egg(egg, ns.target)


if __name__ == "__main__":  # pragma: nocover
    main()



from __future__ import absolute_import

import os.path
import shutil
import tempfile

from egginst.main import EggInst
from egginst.tests.common import DUMMY_EGG, NOSE_1_2_1, NOSE_1_3_0
from egginst.utils import makedirs
from egginst.vendor.six.moves import unittest


from enstaller.egg_meta import split_eggname
from enstaller.errors import EnpkgError, MissingDependency, NoPackageFound
from enstaller.repository import Repository

from ..core import Solver
from ..request import Request
from ..requirement import Requirement

from enstaller.tests.common import (dummy_installed_package_factory,
                                    dummy_repository_package_factory,
                                    repository_factory)


class TestSolverNoDependencies(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_install_simple(self):
        entries = [
            dummy_repository_package_factory("numpy", "1.6.1", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2),
            dummy_repository_package_factory("numpy", "1.7.1", 2),
        ]

        r_actions = [
            ('install', 'numpy-1.8.0-2.egg')
        ]

        repository = repository_factory(entries)
        installed_repository = Repository._from_prefixes([self.prefix])
        solver = Solver(repository, installed_repository)

        request = Request()
        request.install(Requirement("numpy"))
        actions = solver.resolve(request)

        self.assertEqual(actions, r_actions)

    def test_install_no_egg_entry(self):
        # Given
        entries = [
            dummy_repository_package_factory("numpy", "1.6.1", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2),
        ]

        repository = repository_factory(entries)
        installed_repository = Repository._from_prefixes([self.prefix])
        solver = Solver(repository, installed_repository)

        request = Request()
        request.install(Requirement("scipy"))

        # When/Then
        with self.assertRaises(NoPackageFound):
            solver.resolve(request)

    def test_install_missing_dependency(self):
        # Given
        entries = [
            dummy_repository_package_factory("numpy", "1.8.0", 2,
                                             dependencies=["MKL 10.3"]),
        ]

        repository = repository_factory(entries)
        installed_repository = Repository._from_prefixes([self.prefix])
        solver = Solver(repository, installed_repository)

        request = Request()
        request.install(Requirement("numpy"))

        # When/Then
        with self.assertRaises(MissingDependency):
            solver.resolve(request)

    def test_remove_actions(self):
        # Given
        repository = Repository()

        for egg in [DUMMY_EGG]:
            egginst = EggInst(egg, self.prefix)
            egginst.install()

        solver = Solver(repository, Repository._from_prefixes([self.prefix]))

        request = Request()
        request.remove(Requirement("dummy"))

        # When
        actions = solver.resolve(request)

        # Then
        self.assertEqual(actions, [("remove", os.path.basename(DUMMY_EGG))])

    def test_remove_non_existing(self):
        # Given
        entries = [
            dummy_repository_package_factory("numpy", "1.6.1", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2),
        ]

        repository = repository_factory(entries)
        solver = Solver(repository, Repository._from_prefixes([self.prefix]))

        request = Request()
        request.remove(Requirement("numpy"))

        # When/Then
        with self.assertRaises(EnpkgError):
            solver.resolve(request)

    def test_chained_override_update(self):
        """ Test update to package with latest version in lower prefix
        but an older version in primary prefix.
        """
        # Given
        l0_egg = NOSE_1_3_0
        l1_egg = NOSE_1_2_1

        expected_actions = [
            ('remove', os.path.basename(l1_egg)),
            ('install', os.path.basename(l0_egg)),
        ]

        entries = [
            dummy_repository_package_factory(
                *split_eggname(os.path.basename(l0_egg))),
        ]

        repository = repository_factory(entries)

        l0 = os.path.join(self.prefix, 'l0')
        l1 = os.path.join(self.prefix, 'l1')
        makedirs(l0)
        makedirs(l1)

        # Install latest version in l0
        EggInst(l0_egg, l0).install()
        # Install older version in l1
        EggInst(l1_egg, l1).install()

        repository = repository_factory(entries)
        installed_repository = Repository._from_prefixes([l1])
        solver = Solver(repository, installed_repository)

        request = Request()
        request.install(Requirement("nose"))

        # When
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions, expected_actions)


class TestSolverDependencies(unittest.TestCase):
    def test_simple(self):
        # Given
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2,
                                             dependencies=["MKL 10.3"]),
        ]

        repository = repository_factory(entries)
        installed_repository = Repository()

        expected_actions = [
            ('install', "MKL-10.3-1.egg"),
            ('install', "numpy-1.8.0-2.egg"),
        ]

        request = Request()
        request.install(Requirement("numpy"))

        # When
        solver = Solver(repository, installed_repository)
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions, expected_actions)

    def test_simple_installed(self):
        # Given
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2,
                                             dependencies=["MKL 10.3"]),
        ]

        repository = repository_factory(entries)
        installed_repository = Repository()
        installed_repository.add_package(
            dummy_installed_package_factory("MKL", "10.3", 1)
        )

        expected_actions = [
            ('install', "numpy-1.8.0-2.egg"),
        ]

        # When
        request = Request()
        request.install(Requirement("numpy"))

        solver = Solver(repository, installed_repository)
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions, expected_actions)

    def test_simple_all_installed(self):
        # Given
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 2,
                                             dependencies=["MKL 10.3"]),
        ]

        repository = repository_factory(entries)

        installed_entries = [
            dummy_installed_package_factory("MKL", "10.3", 1),
            dummy_installed_package_factory("numpy", "1.8.0", 2)
        ]
        installed_repository = Repository()
        for package in installed_entries:
            installed_repository.add_package(package)

        # When
        request = Request()
        request.install(Requirement("numpy"))

        solver = Solver(repository, installed_repository)
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions, [])

        # When
        solver = Solver(repository, installed_repository, force=True)
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions,
                             [("remove", "numpy-1.8.0-2.egg"),
                              ("install", "numpy-1.8.0-2.egg")])

        # When
        solver = Solver(repository, installed_repository, force=True,
                        forceall=True)
        actions = solver.resolve(request)

        # Then
        self.assertListEqual(actions,
                             [("remove", "numpy-1.8.0-2.egg"),
                              ("remove", "MKL-10.3-1.egg"),
                              ("install", "MKL-10.3-1.egg"),
                              ("install", "numpy-1.8.0-2.egg")])



from __future__ import print_function

import logging
import os
import sys
import re
from os.path import abspath, basename, join, isdir, isfile, islink

from egginst.exe_data import cli, gui
from egginst.utils import on_win, rm_rf

hashbang_pat = re.compile(r'#!.+$', re.M)
executable = sys.executable
logger = logging.getLogger(__name__)


def get_executable(pythonw=False, with_quotes=False):
    res = executable
    if on_win:
        # sys.executable may actually be pythonw.exe in order to avoid
        # popping up a cmd shell during install.
        p = re.compile(r'pythonw?\.exe', re.I)
        res = p.sub('pythonw.exe' if pythonw else 'python.exe', res)
    if with_quotes:
        res = '"%s"' % res
    return res


def write_exe(dst, script_type='console_scripts'):
    """
    This function is only used on Windows.   It either writes cli.exe or
    gui.exe to the destination specified, depending on script_type.

    The binary content of these files are read from the module exe_data,
    which may be generated by the following small script (run from the
    setuptools directory which contains the file cli.exe and gui.exe:
    fo = open('exe_data.py', 'w')
    for name in ['cli', 'gui']:
        data = open('%s.exe' % name, 'rb').read()
        fo.write('%s = %r\n' % (name, data))
    fo.close()
    """
    if script_type == 'console_scripts':
        data = cli
    elif script_type == 'gui_scripts':
        data = gui
    else:
        raise Exception("Did not except script_type=%r" % script_type)

    rm_rf(dst)
    try:
        with open(dst, 'wb') as fp:
            fp.write(data)
    except IOError:
        # When bootstrapping, the file egginst.exe is in use and can therefore
        # not be rewritten, which is OK since its content is always the same.
        pass
    os.chmod(dst, 0o755)


def create_proxy(src, bin_dir):
    """
    create a proxy of src in bin_dir (Windows only)
    """
    logger.info("Creating proxy executable to: %r", src)
    assert src.endswith('.exe')

    dst_name = basename(src)
    if dst_name.startswith('epd-'):
        dst_name = dst_name[4:]

    dst = join(bin_dir, dst_name)
    write_exe(dst)

    dst_script = dst[:-4] + '-script.py'
    rm_rf(dst_script)

    with open(dst_script, 'w') as fo:
        fo.write('''\
#!"%(python)s"
# This proxy was created by egginst from an egg with special instructions
#
import sys
import subprocess

src = %(src)r

sys.exit(subprocess.call([src] + sys.argv[1:]))
''' % dict(python=get_executable(), src=src))
    return dst, dst_script


def create_proxies(egg):
    # This function is called on Windows only
    if not isdir(egg.bin_dir):
        os.makedirs(egg.bin_dir)

    for line in egg.iter_files_to_install():
        arcname, action = line.split()
        logger.info("arcname=%r    action=%r", arcname, action)

        if action == 'PROXY':
            ei = 'EGG-INFO/'
            if arcname.startswith(ei):
                src = abspath(join(egg.meta_dir, arcname[len(ei):]))
            else:
                src = abspath(join(egg.prefix, arcname))
            logger.info("     src: %r", src)
            egg.files.extend(create_proxy(src, egg.bin_dir))
        else:
            data = egg.z.read(arcname)
            dst = abspath(join(egg.prefix, action, basename(arcname)))
            logger.info("     dst: %r", dst)
            rm_rf(dst)
            with open(dst, 'wb') as fo:
                fo.write(data)
            egg.files.append(dst)


def write_script(path, entry_pt, egg_name):
    """
    Write an entry point script to path.
    """
    logger.info('Creating script: %s', path)

    assert entry_pt.count(':') == 1
    module, func = entry_pt.strip().split(':')

    rm_rf(path)
    with open(path, 'w') as fo:
        fo.write('''\
#!%(python)s
# This script was created by egginst when installing:
#
#   %(egg_name)s
#
if __name__ == '__main__':
    import sys
    from %(module)s import %(func)s

    sys.exit(%(func)s())
''' % dict(python=get_executable(pythonw=path.endswith('.pyw'),
                                 with_quotes=on_win),
           egg_name=egg_name, module=module, func=func))
    os.chmod(path, 0o755)


def create(egg, conf):
    if not isdir(egg.bin_dir):
        os.makedirs(egg.bin_dir)

    for script_type in ['gui_scripts', 'console_scripts']:
        if script_type not in conf.sections():
            continue
        for name, entry_pt in conf.items(script_type):
            fname = name
            if on_win:
                exe_path = join(egg.bin_dir, '%s.exe' % name)
                write_exe(exe_path, script_type)
                egg.files.append(exe_path)
                fname += '-script.py'
                if script_type == 'gui_scripts':
                    fname += 'w'
            path = join(egg.bin_dir, fname)
            write_script(path, entry_pt, egg.fn)
            egg.files.append(path)


def fix_script(path):
    """
    Fixes a single located at path.
    """
    if islink(path) or not isfile(path):
        return

    with open(path, "rb") as fi:
        data = fi.read()

    try:
        data = data.decode("utf8")
    except UnicodeDecodeError:
        return

    if ' egginst ' in data:
        # This string is in the comment when write_script() creates
        # the script, so there is no need to fix anything.
        return

    m = hashbang_pat.match(data)
    if not (m and 'python' in m.group().lower()):
        return

    python = get_executable(with_quotes=on_win)
    new_data = hashbang_pat.sub('#!' + python.replace('\\', '\\\\'),
                                data, count=1)
    if new_data == data:
        return
    logger.info("Updating: %r", path)

    with open(path, 'wb') as fo:
        fo.write(new_data.encode("utf8"))

    os.chmod(path, 0o755)


def fix_scripts(egg):
    for path in egg.files:
        if path.startswith(egg.bin_dir):
            fix_script(path)



import unittest

from egginst.eggmeta import info_from_z, parse_rawspec
from egginst.vendor.six import BytesIO
from egginst._zipfile import ZipFile

NUMPY_1_4_0_WIN32 = """\
metadata_version = '1.1'
name = 'numpy'
version = '1.4.0'
build = 3

arch = 'x86'
platform = 'win32'
osdist = None
python = '2.6'
packages = []
"""

PIL_1_1_7_WIN32 = """\
metadata_version = '1.1'
name = 'PIL'
version = '1.1.7'
build = 3

arch = 'x86'
platform = 'win32'
osdist = None
python = '2.7'
packages = []
"""


def _create_inmemory_egg(archives):
    s = BytesIO()
    with ZipFile(s, "w") as z:
        for arcname, data in archives.items():
            z.writestr(arcname, data)
    s.seek(0)

    return s


class TestParseRawSpec(unittest.TestCase):
    def test_simple(self):
        data = NUMPY_1_4_0_WIN32

        r_metadata = {
            "name": "numpy",
            "version": "1.4.0",
            "build": 3,
            "arch": "x86",
            "platform": "win32",
            "osdist": None,
            "python": "2.6",
            "packages": []
        }
        metadata = parse_rawspec(data)

        self.assertEqual(metadata, r_metadata)


class TestInfoFromZ(unittest.TestCase):
    def test_simple(self):
        r_metadata = {
            "name": "numpy",
            "version": "1.4.0",
            "build": 3,
            "arch": "x86",
            "platform": "win32",
            "osdist": None,
            "python": "2.6",
            "packages": [],
            "type": "egg",
        }

        s = _create_inmemory_egg({"EGG-INFO/spec/depend": NUMPY_1_4_0_WIN32})
        with ZipFile(s) as z:
            metadata = info_from_z(z)
            self.assertEqual(metadata, r_metadata)

    def test_name_casing(self):
        r_metadata = {
            "name": "pil",
            "version": "1.1.7",
            "build": 3,
            "arch": "x86",
            "platform": "win32",
            "osdist": None,
            "python": "2.7",
            "packages": [],
            "type": "egg",
        }

        s = _create_inmemory_egg({"EGG-INFO/spec/depend": PIL_1_1_7_WIN32})
        with ZipFile(s) as z:
            metadata = info_from_z(z)
            self.assertEqual(metadata, r_metadata)



from __future__ import print_function

import sys
from os.path import isfile, join


def main(prefix=sys.prefix, hook=False, pkgs_dir=None, verbose=False):
    """
    To bootstrap enstaller into a Python environment, used the following
    code:

    sys.path.insert(0, '/path/to/enstaller.egg')
    from egginst.bootstrap import main
    main()
    """
    import egginst

    # This is the path to the egg which we want to install.
    # Note that whoever calls this function has inserted the egg to the
    # from of sys.path
    egg_path = sys.path[0]

    print("Bootstrapping:", egg_path)
    ei = egginst.EggInst(egg_path, prefix,
                         hook=hook, pkgs_dir=pkgs_dir)
    ei.install()


def fix_easy_pth(pth):
    new_lines = []
    needs_rewrite = False
    for line in open(pth):
        line = line.strip()
        if 'enstaller' in line.lower():
            needs_rewrite = True
        else:
            new_lines.append(line)

    if needs_rewrite:
        fo = open(pth, 'w')
        for line in new_lines:
            fo.write(line + '\n')
        fo.close()
        print("Removed enstaller entry from", pth)


def remove_and_fix():
    # Remove and fix some files in site-packages
    from egginst.utils import rel_site_packages

    site_dir = join(sys.prefix, rel_site_packages)

    # If there an easy-install.pth in site-packages, remove and
    # occurrences of enstaller from it.
    pth = join(site_dir, 'easy-install.pth')
    if isfile(pth):
        fix_easy_pth(pth)


def cli():
    """
    CLI (for executable egg)
    """
    from optparse import OptionParser
    from enstaller import __version__

    p = OptionParser(usage="<executable egg> [options]",
                     description="bootstraps enstaller %(__version__)s into "
                                 "the current Python environment" % locals())

    p.add_option("--hook",
                 action="store_true")
    p.add_option("--prefix",
                 action="store",
                 default=sys.prefix,
                 help="install prefix, defaults to %default")
    p.add_option('-v', "--verbose", action="store_true")
    p.add_option('--version', action="store_true")

    opts, args = p.parse_args()

    if opts.version:
        print("enstaller version:", __version__)
        return

    main(opts.prefix, opts.hook, opts.verbose)

    remove_and_fix()



from __future__ import print_function

import contextlib
import os.path
import random
import sys

import mock

from egginst.main import name_version_fn
from egginst.tests.common import DUMMY_EGG_SIZE, DUMMY_EGG, \
    DUMMY_EGG_MTIME, DUMMY_EGG_MD5
from egginst.vendor.six.moves import unittest

from enstaller.utils import canonical, comparable_version, input_auth, \
    path_to_uri, uri_to_path, info_file, cleanup_url, \
    prompt_yes_no, under_venv, real_prefix
from .common import mock_input, mock_print, mock_raw_input


class TestUtils(unittest.TestCase):

    def test_canonical(self):
        for name, cname in [
            ('NumPy', 'numpy'),
            ('MySql-python', 'mysql_python'),
            ('Python-dateutil', 'python_dateutil'),
        ]:
            self.assertEqual(canonical(name), cname)

    def test_naming(self):
        for fn, name, ver, cname in [
            ('NumPy-1.5-py2.6-win32.egg', 'NumPy', '1.5-py2.6-win32', 'numpy'),
            ('NumPy-1.5-2.egg', 'NumPy', '1.5-2', 'numpy'),
            ('NumPy-1.5.egg', 'NumPy', '1.5', 'numpy'),
        ]:
            self.assertEqual(name_version_fn(fn), (name, ver))
            self.assertEqual(name.lower(), cname)
            self.assertEqual(canonical(name), cname)

    def test_comparable_version(self):
        for versions in (
            ['1.0.4', '1.2.1', '1.3.0b1', '1.3.0', '1.3.10',
             '1.3.11.dev7', '1.3.11.dev12', '1.3.11.dev111',
             '1.3.11', '1.3.143',
             '1.4.0.dev7749', '1.4.0rc1', '1.4.0rc2', '1.4.0'],
            ['2008j', '2008k', '2009b', '2009h', '2010b'],
            ['0.99', '1.0a2', '1.0b1', '1.0rc1', '1.0', '1.0.1'],
            ['2.0.8', '2.0.10', '2.0.10.1', '2.0.11'],
            ['0.10.1', '0.10.2', '0.11.dev1324', '0.11'],
        ):
            org = list(versions)
            random.shuffle(versions)
            versions.sort(key=comparable_version)
            self.assertEqual(versions, org)

    def test_info_file(self):
        r_info = {
            "size": DUMMY_EGG_SIZE,
            "mtime": DUMMY_EGG_MTIME,
            "md5": DUMMY_EGG_MD5
        }

        info = info_file(DUMMY_EGG)
        self.assertEqual(info, r_info)

    def test_cleanup_url(self):
        r_data = [
            ("http://www.acme.com/", "http://www.acme.com/"),
            ("http://www.acme.com", "http://www.acme.com/"),
            ("file:///foo/bar", "file:///foo/bar/"),
        ]

        for url, r_url in r_data:
            self.assertEqual(cleanup_url(url), r_url)

    def test_cleanup_url_dir(self):
        if sys.platform == "win32":
            # The \\ -> / replace is a hack to deal with expanduser being
            # inconsistent on windows depending on whether HOME, USERPROFILE or
            # HOMEPATH is used
            home = os.path.expanduser("~").replace("\\", "/")
            p = "/" + os.path.abspath(home)
        else:
            p = os.path.abspath(os.path.expanduser("~"))
        r_url = "file://{0}/".format(p)

        url = "~"

        self.assertEqual(cleanup_url(url), r_url)

    def test_cleanup_url_relative_path(self):
        url, r_url = "file://foo/bar", "file://foo/bar/"

        self.assertEqual(cleanup_url(url), r_url)


class TestUri(unittest.TestCase):
    def test_path_to_uri_simple(self):
        """Ensure path to uri conversion works."""
        # XXX: this is a bit ugly, but urllib does not allow to select which OS
        # we want (there is no 'nturllib' or 'posixurllib' as there is for path.
        if sys.platform == "win32":
            r_uri = "file:///C:/Users/vagrant/yo"
            uri = path_to_uri("C:\\Users\\vagrant\\yo")
        else:
            r_uri = "file:///home/vagrant/yo"
            uri = path_to_uri("/home/vagrant/yo")
        self.assertEqual(r_uri, uri)

    def test_uri_to_path_simple(self):
        if sys.platform == "win32":
            r_path = "C:\\Users\\vagrant\\yo"
            uri = "file:///C:/Users/vagrant/yo"
        else:
            r_path = "/home/vagrant/yo"
            uri = "file:///home/vagrant/yo"

        path = uri_to_path(uri)
        self.assertEqual(r_path, path)

    def test_uri_to_path_simple_local(self):
        if sys.platform == "win32":
            r_path = "vagrant\\yo"
            uri = "file://vagrant/yo"
        else:
            r_path = "vagrant/yo"
            uri = "file://vagrant/yo"

        path = uri_to_path(uri)
        self.assertEqual(r_path, path)

    @contextlib.contextmanager
    def _patch_prefix(self, attr_name, prefix_value):
        # attr_name can be None to remove all venv prefixes
        prefix_names = ("real_prefix", "base_prefix")
        orig_attrs = {}
        for prefix_name in prefix_names:
            if hasattr(sys, prefix_name):
                orig_attrs[prefix_name] = getattr(sys, prefix_name)
                delattr(sys, prefix_name)

        try:
            if attr_name is not None:
                with mock._patch_object(sys, attr_name, prefix_value,
                                        create=True):
                    yield
            else:
                yield
        finally:
            for prefix_name, prefix_value in orig_attrs.items():
                setattr(sys, prefix_name, prefix_value)

    def _mock_under_non_venv(self):
        return self._patch_prefix(None, None)

    def _mock_under_venv(self):
        return self._patch_prefix("base_prefix",
                                  os.path.join(sys.prefix, 'venv'))

    def _mock_under_virtualenv(self):
        return self._patch_prefix("real_prefix",
                                  os.path.join(sys.prefix, 'venv'))

    def _mock_under_base_venv(self):
        return self._patch_prefix("base_prefix", sys.prefix)

    def _mock_under_virtualenv_under_base_venv(self):
        with self._patch_prefix("base_prefix", sys.prefix):
            return self._mock_under_virtualenv()

    def test_under_venv(self):
        def _check(mock_method, is_under_venv, prefix_name):
            with mock_method():
                self.assertEqual(under_venv(), is_under_venv)
                self.assertEqual(real_prefix(), getattr(sys, prefix_name))

        _check(self._mock_under_non_venv, False, "prefix")

        _check(self._mock_under_venv, True, "base_prefix")

        _check(self._mock_under_virtualenv, True, "real_prefix")

        _check(self._mock_under_base_venv, False, "prefix")

        _check(self._mock_under_virtualenv_under_base_venv,
               True, "real_prefix")


class TestPromptYesNo(unittest.TestCase):
    def test_simple(self):
        # Given
        message = "Do you want to do it ?"

        # When
        with mock_print() as m:
            with mock_raw_input("yes"):
                res = prompt_yes_no(message)

        # Then
        self.assertEqual(m.value.rstrip(), message)
        self.assertTrue(res)

    def test_simple_no(self):
        # Given
        message = "Do you want to do it ?"

        # When
        with mock_print() as m:
            with mock_raw_input("no"):
                res = prompt_yes_no(message)

        # Then
        self.assertEqual(m.value.rstrip(), message)
        self.assertFalse(res)

    def test_simple_force_yes(self):
        # Given
        message = "Do you want to do it ?"

        # When
        with mock_print() as m:
            with mock_raw_input("yes") as mocked_input:
                res = prompt_yes_no(message, True)

        # Then
        self.assertEqual(m.value.rstrip(), message)
        self.assertTrue(res)
        mocked_input.assert_called()


FAKE_USER = "john.doe"
FAKE_PASSWORD = "fake_password"


class TestInputAuth(unittest.TestCase):
    @mock.patch("enstaller.utils.getpass.getpass", lambda ignored: FAKE_PASSWORD)
    def test_simple(self):
        with mock_input(FAKE_USER):
            self.assertEqual(input_auth(), (FAKE_USER, FAKE_PASSWORD))

    @mock.patch("enstaller.utils.getpass.getpass", lambda ignored: FAKE_PASSWORD)
    def test_empty(self):
        with mock_input(""):
            self.assertEqual(input_auth(), (None, None))



import os
import shutil
import tempfile

from egginst._compat import assertCountEqual
from egginst.testing_utils import ControlledEnv, assert_same_fs
from egginst.vendor.six.moves import unittest


class TestControlledEnv(unittest.TestCase):
    def test_simple(self):
        # Given
        env = ControlledEnv()

        # When
        env["key"] = "value"

        # Then
        self.assertEqual(env["key"], "value")

    def test_keys(self):
        # Given
        env = ControlledEnv()

        # When
        env["key1"] = "value1"
        env["key2"] = "value2"

        # Then
        assertCountEqual(self, env.keys(), list(os.environ.keys()) + ["key1", "key2"])

    def test_get_item(self):
        # Given
        env = ControlledEnv(["ignored_key"])

        # When/Then
        with self.assertRaises(KeyError):
            env["ignored_key"]

    def test_custom_env(self):
        # Given
        env = ControlledEnv(environ={"key1": "value1"})

        # When/Then
        self.assertFalse(list(os.environ.keys())[0] in env)


class TestAssertSameFs(unittest.TestCase):
    def setUp(self):
        self.scratch = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.scratch)

    def test_empty(self):
        # When/Then
        with assert_same_fs(self, self.scratch):
            pass

    def test_simple(self):
        # When/Then
        with assert_same_fs(self, self.scratch):
            path = os.path.join(self.scratch, "yo")
            with open(path, "w") as fp:
                fp.write("")
            os.unlink(path)

        # When/Then
        with self.assertRaises(AssertionError):
            with assert_same_fs(self, self.scratch):
                path = os.path.join(self.scratch, "yo")
                with open(path, "w") as fp:
                    fp.write("")

    def test_remove_file(self):
        # Given
        path = os.path.join(self.scratch, "yo")
        with open(path, "w") as fp:
            fp.write("")

        # When/Then
        with assert_same_fs(self, self.scratch):
            self.assertTrue(os.path.exists(path))



from __future__ import absolute_import

import abc

from egginst._compat import with_metaclass
from egginst.vendor.six.moves import urllib

from enstaller.errors import AuthFailedError
from enstaller.vendor import requests

from .user_info import UserInfo


class IAuthManager(with_metaclass(abc.ABCMeta)):
    @abc.abstractproperty
    def authenticate(self, session, auth):
        """ Authenticate.

        Parameters
        ----------
        session : Session
            The connection handled used to manage http connections(s).
        auth : tuple
            The (username, password) pair for authentication.

        Raises
        ------
        an AuthFailedError if authentication failed.
        """


class LegacyCanopyAuthManager(object):
    @classmethod
    def from_configuration(cls, configuration):
        """ Create a LegacyCanopyAuthManager instance from an enstaller config
        object.
        """
        return cls(configuration.api_url)

    def __init__(self, url):
        self.url = url
        self._auth = None

    def authenticate(self, session, auth):
        try:
            resp = session._raw_get(self.url, auth=auth)
        except requests.exceptions.SSLError as e:
            raise
        except requests.exceptions.ConnectionError as e:
            raise AuthFailedError(str(e), e)

        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise AuthFailedError("Authentication error: %r" % str(e), e)

        # See if web API refused to authenticate
        user = UserInfo.from_json_string(resp.content.decode("utf8"))
        if not user.is_authenticated:
            msg = 'Authentication error: Invalid user login.'
            raise AuthFailedError(msg)

        self._auth = auth


class OldRepoAuthManager(object):
    @classmethod
    def from_configuration(cls, configuration):
        """ Create a OldRepoAuthManager instance from an enstaller config
        object.
        """
        return cls(configuration.indices)

    def __init__(self, index_urls):
        self.index_urls = index_urls
        self._auth = None

    def authenticate(self, session, auth):
        for index_url, _ in self.index_urls:
            parse = urllib.parse.urlparse(index_url)
            if parse.scheme in ("http", "https"):
                try:
                    resp = session._raw_head(index_url, auth=auth)
                except requests.exceptions.SSLError as e:
                    raise
                except requests.exceptions.ConnectionError as e:
                    raise AuthFailedError(str(e), e)

                try:
                    resp.raise_for_status()
                except requests.exceptions.HTTPError as e:
                    http_code = resp.status_code
                    if http_code in (401, 403):
                        msg = "Authentication error: {0!r}".format(str(e))
                        raise AuthFailedError(msg, e)
                    elif http_code == 404:
                        msg = "Could not access repo {0!r} (error: {1!r})". \
                              format(index_url, str(e))
                        raise AuthFailedError(msg, e)
                    else:
                        raise AuthFailedError(str(e), e)
        self._auth = auth


class BroodBearerTokenAuth(requests.auth.AuthBase):

    def __init__(self, token):
        self._token = token

    def __call__(self, request):
        request.headers['Authorization'] = 'Bearer {0}'.format(self._token)
        return request


class BroodAuthenticator(object):
    """ Token-based authenticator for brood stores."""
    @classmethod
    def from_configuration(cls, configuration):
        """ Create a BroodAuthenticator instance from an enstaller config
        object.
        """
        return cls(configuration.store_url)

    def __init__(self, url):
        self.url = url
        self._auth = None

    def authenticate(self, session, auth):
        url = self.url + "/api/v0/json/auth/tokens/auth"
        try:
            resp = session._raw_post(url, auth=auth)
        except requests.exceptions.ConnectionError as e:
            raise AuthFailedError(e)

        try:
            resp.raise_for_status()
        except requests.exceptions.HTTPError as e:
            raise AuthFailedError("Authentication error: %r" % str(e))

        token = resp.json()["token"]
        self._auth = BroodBearerTokenAuth(token)


IAuthManager.register(LegacyCanopyAuthManager)
IAuthManager.register(OldRepoAuthManager)
IAuthManager.register(BroodAuthenticator)



from __future__ import absolute_import

import os

from ..py27compat import unicode_str
from ..backend import KeyringBackend
from ..errors import PasswordDeleteError
from ..errors import PasswordSetError, ExceptionRaisedContext
from ..util import properties
from ..util import XDG

try:
    from PyKDE4.kdeui import KWallet
    from PyQt4 import QtGui
except ImportError:
    pass

kwallet = None

def open_kwallet(kwallet_module=None, qt_module=None):

    # If we specified the kwallet_module and/or qt_module, surely we won't need
    # the cached kwallet object...
    if kwallet_module is None and qt_module is None:
        global kwallet
        if not kwallet is None:
            return kwallet

    # Allow for the injection of module-like objects for testing purposes.
    if kwallet_module is None:
        kwallet_module = KWallet.Wallet
    if qt_module is None:
        qt_module = QtGui

    # KDE wants us to instantiate an application object.
    app = None
    if qt_module.qApp.instance() == None:
        app = qt_module.QApplication([])
    try:
        window = qt_module.QWidget()
        kwallet = kwallet_module.openWallet(
            kwallet_module.NetworkWallet(),
            window.winId(),
            kwallet_module.Synchronous)
        if kwallet is not None:
            if not kwallet.hasFolder('Python'):
                kwallet.createFolder('Python')
            kwallet.setFolder('Python')
            return kwallet
    finally:
        if app:
            app.exit()

class Keyring(KeyringBackend):
    """KDE KWallet"""

    @properties.ClassProperty
    @classmethod
    @XDG.Preference('KDE')
    def priority(cls):
        with ExceptionRaisedContext() as exc:
            KWallet.__name__
        if exc:
            raise RuntimeError("KDE libraries not available")
        # Infer if KDE environment is active based on environment vars.
        # TODO: Does PyKDE provide a better indicator?
        kde_session_keys = (
            'KDE_SESSION_ID', # most environments
            'KDE_FULL_SESSION', # openSUSE
        )
        if not set(os.environ).intersection(kde_session_keys):
            return 0
        return 5

    def get_password(self, service, username):
        """Get password of the username for the service
        """
        key = username + '@' + service
        network = KWallet.Wallet.NetworkWallet()
        wallet = open_kwallet()
        if wallet is None:
            # the user pressed "cancel" when prompted to unlock their keyring.
            return None
        if wallet.keyDoesNotExist(network, 'Python', key):
            return None

        result = wallet.readPassword(key)[1]
        # The string will be a PyQt4.QtCore.QString, so turn it into a unicode
        # object.
        return unicode_str(result)

    def set_password(self, service, username, password):
        """Set password for the username of the service
        """
        wallet = open_kwallet()
        if wallet is None:
            # the user pressed "cancel" when prompted to unlock their keyring.
            raise PasswordSetError("Cancelled by user")
        wallet.writePassword(username+'@'+service, password)

    def delete_password(self, service, username):
        """Delete the password for the username of the service.
        """
        key = username + '@' + service
        wallet = open_kwallet()
        if wallet is None:
            # the user pressed "cancel" when prompted to unlock their keyring.
            raise PasswordDeleteError("Cancelled by user")
        if wallet.keyDoesNotExist(wallet.walletName(), 'Python', key):
            raise PasswordDeleteError("Password not found")
        wallet.removeEntry(key)



import operator
import os.path
import time

from egginst._compat import assertCountEqual
from egginst.main import EggInst
from egginst.utils import compute_md5
from egginst.testing_utils import slow
from egginst.tests.common import _EGGINST_COMMON_DATA, DUMMY_EGG, create_venv, mkdtemp
from egginst.vendor.six.moves import unittest

from enstaller.compat import path_to_uri
from enstaller.errors import NoSuchPackage
from enstaller.versions.enpkg import EnpkgVersion

from enstaller.repository import (InstalledPackageMetadata, PackageMetadata,
                                  Repository, RepositoryPackageMetadata,
                                  egg_name_to_name_version)
from enstaller.solver import Requirement
from enstaller.tests.common import dummy_installed_package_factory
from enstaller.utils import PY_VER


class TestEggNameToNameVersion(unittest.TestCase):
    def test_simple(self):
        # given
        egg_name = "numpy-1.8.0-1.egg"

        # When
        name, version = egg_name_to_name_version(egg_name)

        # Then
        self.assertEqual(name, "numpy")
        self.assertEqual(version, "1.8.0-1")

    def test_simple_uppercase(self):
        # given
        egg_name = "MKL-10.3-1.egg"

        # When
        name, version = egg_name_to_name_version(egg_name)

        # Then
        self.assertEqual(name, "mkl")
        self.assertEqual(version, "10.3-1")

    def test_invalid(self):
        # given
        egg_name = "nono"

        # When
        with self.assertRaises(ValueError):
            egg_name_to_name_version(egg_name)


class TestPackage(unittest.TestCase):
    def test_repr(self):
        # Given
        version = EnpkgVersion.from_string("1.3.0-1")
        metadata = PackageMetadata("nose-1.3.0-1.egg", "nose", version, [],
                                   "2.7")

        # When
        r = repr(metadata)

        # Then
        self.assertEqual(r, "PackageMetadata('nose-1.3.0-1', key='nose-1.3.0-1.egg')")

    def test_from_egg(self):
        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")

        # When
        metadata = PackageMetadata.from_egg(path)

        # Then
        self.assertEqual(metadata.name, "nose")
        self.assertEqual(metadata.version, EnpkgVersion.from_string("1.3.0-1"))


class TestRepositoryPackage(unittest.TestCase):
    def test_s3index_data(self):
        # Given
        md5 = "c68bb183ae1ab47b6d67ca584957c83c"
        r_s3index_data = {
            "available": True,
            "build": 1,
            "md5": md5,
            "mtime": 0.0,
            "name": "nose",
            "packages": [],
            "product": "free",
            "python": "2.7",
            "size": 1,
            "type": "egg",
            "version": "1.3.0",

        }
        version = EnpkgVersion.from_string("1.3.0-1")
        metadata = RepositoryPackageMetadata("nose-1.3.0-1.egg", "nose",
                                             version, [], "2.7", 1, md5,
                                             0.0, "free", True, "")

        # When/Then
        self.assertEqual(metadata.s3index_data, r_s3index_data)

    def test_from_egg(self):
        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")

        # When
        metadata = RepositoryPackageMetadata.from_egg(path)

        # Then
        self.assertEqual(metadata.name, "nose")
        self.assertEqual(metadata.version, EnpkgVersion.from_string("1.3.0-1"))
        self.assertEqual(metadata.store_location,
                         "{0}/".format(path_to_uri(_EGGINST_COMMON_DATA)))
        self.assertEqual(metadata.source_url, path_to_uri(path))

    def test_repr(self):
        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")
        store_location = path_to_uri(os.path.dirname(path)) + "/"
        r_repr = ("RepositoryPackageMetadata('nose-1.3.0-1', "
                  "key='nose-1.3.0-1.egg', available=True, product=None, "
                  "store_location='{0}')".format(store_location))

        # When
        metadata = RepositoryPackageMetadata.from_egg(path)

        # Then
        self.assertEqual(repr(metadata), r_repr)


class TestInstalledPackage(unittest.TestCase):
    def test_from_meta_dir(self):
        # Given
        json_dict = {
            "arch": "amd64",
            "build": 1,
            "ctime": "Thu Apr 24 15:41:24 2014",
            "hook": False,
            "key": "VTK-5.10.1-1.egg",
            "name": "vtk",
            "osdist": "RedHat_5",
            "packages": [],
            "platform": "linux2",
            "python": "2.7",
            "type": "egg",
            "version": "5.10.1"
        }

        # When
        metadata = InstalledPackageMetadata.from_installed_meta_dict(json_dict)

        # Then
        self.assertEqual(metadata.key, "VTK-5.10.1-1.egg")

    def test_from_meta_dir_no_packages(self):
        # Given
        json_dict = {
            "arch": "amd64",
            "build": 1,
            "ctime": "Thu Apr 24 15:41:24 2014",
            "hook": False,
            "key": "VTK-5.10.1-1.egg",
            "name": "vtk",
            "osdist": "RedHat_5",
            "platform": "linux2",
            "python": "2.7",
            "type": "egg",
            "version": "5.10.1"
        }

        # When
        metadata = InstalledPackageMetadata.from_installed_meta_dict(json_dict)

        # Then
        self.assertEqual(metadata.key, "VTK-5.10.1-1.egg")
        self.assertEqual(metadata.packages, [])

    def test_from_old_meta_dir(self):
        # Given
        json_dict = {
            "build": 1,
            "hook": False,
            "key": "appinst-2.1.2-1.egg",
            "name": "appinst",
            "version": "2.1.2"
        }

        # When
        metadata = InstalledPackageMetadata.from_installed_meta_dict(json_dict)

        # Then
        self.assertEqual(metadata.key, "appinst-2.1.2-1.egg")
        self.assertEqual(metadata.python, PY_VER)
        self.assertEqual(metadata.packages, [])
        self.assertEqual(metadata.ctime, time.ctime(0.0))


class TestRepository(unittest.TestCase):
    def setUp(self):
        eggs = [
            "dummy-1.0.1-1.egg",
            "dummy_with_appinst-1.0.0-1.egg",
            "dummy_with_entry_points-1.0.0-1.egg",
            "dummy_with_proxy-1.3.40-3.egg",
            "dummy_with_proxy_scripts-1.0.0-1.egg",
            "dummy_with_proxy_softlink-1.0.0-1.egg",
            "nose-1.2.1-1.egg",
            "nose-1.3.0-1.egg",
            "nose-1.3.0-2.egg",
        ]
        self.repository = Repository()
        for egg in eggs:
            path = os.path.join(_EGGINST_COMMON_DATA, egg)
            package = RepositoryPackageMetadata.from_egg(path)
            self.repository.add_package(package)

    def test_ctor(self):
        # When
        repository = Repository()

        # Then
        self.assertEqual(len(repository), 0)

        # Given
        eggs = [
            "dummy-1.0.1-1.egg",
            "dummy_with_appinst-1.0.0-1.egg",
            "dummy_with_entry_points-1.0.0-1.egg",
            "dummy_with_proxy-1.3.40-3.egg",
            "dummy_with_proxy_scripts-1.0.0-1.egg",
            "dummy_with_proxy_softlink-1.0.0-1.egg",
            "nose-1.2.1-1.egg",
            "nose-1.3.0-1.egg",
            "nose-1.3.0-2.egg",
        ]
        paths = (os.path.join(_EGGINST_COMMON_DATA, egg) for egg in eggs)
        packages = [RepositoryPackageMetadata.from_egg(path) for path in paths]

        # When
        repository = Repository(packages)

        # Then
        self.assertEqual(len(repository), len(eggs))

    def test_find_package(self):
        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")

        # When
        metadata = self.repository.find_package("nose", "1.3.0-1")

        # Then
        self.assertEqual(metadata.key, "nose-1.3.0-1.egg")

        self.assertEqual(metadata.name, "nose")
        self.assertEqual(metadata.version, EnpkgVersion.from_string("1.3.0-1"))

        self.assertEqual(metadata.packages, [])
        self.assertEqual(metadata.python, "2.7")

        self.assertEqual(metadata.available, True)
        self.assertEqual(metadata.store_location,
                         "{0}/".format(path_to_uri(_EGGINST_COMMON_DATA)))

        self.assertEqual(metadata.size, os.path.getsize(path))
        self.assertEqual(metadata.md5, compute_md5(path))

        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-2.egg")

        # When
        metadata = self.repository.find_package("nose", "1.3.0-2")

        # Then
        self.assertEqual(metadata.key, "nose-1.3.0-2.egg")

        self.assertEqual(metadata.name, "nose")
        self.assertEqual(metadata.version, EnpkgVersion.from_string("1.3.0-2"))

    def test_find_unavailable_package(self):
        # Given/When/Then
        with self.assertRaises(NoSuchPackage):
            self.repository.find_package("nono", "1.4.0-1")

    def test_find_packages(self):
        V = EnpkgVersion.from_string
        # Given/When
        metadata = list(self.repository.find_packages("nose"))
        metadata = sorted(metadata, key=operator.attrgetter("version"))

        # Then
        self.assertEqual(len(metadata), 3)

        self.assertEqual(metadata[0].version, V("1.2.1-1"))
        self.assertEqual(metadata[1].version, V("1.3.0-1"))
        self.assertEqual(metadata[2].version, V("1.3.0-2"))

    def test_find_packages_with_version(self):
        # Given/When
        metadata = list(self.repository.find_packages("nose", "1.3.0-1"))

        # Then
        self.assertEqual(len(metadata), 1)

        self.assertEqual(metadata[0].version,
                         EnpkgVersion.from_string("1.3.0-1"))

    def test_has_package(self):
        # Given
        version = EnpkgVersion.from_string("1.3.0-1")
        available_package = PackageMetadata("nose-1.3.0-1.egg", "nose",
                                            version, [], "2.7")

        version = EnpkgVersion.from_string("1.4.0-1")
        unavailable_package = PackageMetadata("nose-1.4.0-1.egg", "nose",
                                              version, [], "2.7")

        # When/Then
        self.assertTrue(self.repository.has_package(available_package))
        self.assertFalse(self.repository.has_package(unavailable_package))

    def test_iter_most_recent_packages(self):
        # Given
        eggs = ["nose-1.3.0-1.egg", "nose-1.2.1-1.egg"]
        repository = Repository()
        for egg in eggs:
            path = os.path.join(_EGGINST_COMMON_DATA, egg)
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        # When
        metadata = list(repository.iter_most_recent_packages())

        # Then
        self.assertEqual(len(metadata), 1)
        self.assertEqual(metadata[0].version,
                         EnpkgVersion.from_string("1.3.0-1"))

    def test_iter_packages(self):
        # Given
        eggs = ["nose-1.3.0-1.egg", "nose-1.2.1-1.egg"]
        repository = Repository()
        for egg in eggs:
            path = os.path.join(_EGGINST_COMMON_DATA, egg)
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        # When
        metadata = list(repository.iter_packages())

        # Then
        self.assertEqual(len(metadata), 2)
        self.assertEqual(set(m.version for m in metadata),
                         set([EnpkgVersion.from_string("1.2.1-1"),
                              EnpkgVersion.from_string("1.3.0-1")]))

    @slow
    def test_from_prefix(self):
        # Given
        path = DUMMY_EGG
        with mkdtemp() as tempdir:
            create_venv(tempdir)
            installer = EggInst(path, prefix=tempdir)
            installer.install()

            # When
            repository = Repository._from_prefixes([tempdir])

            # Then
            packages = repository.find_packages("dummy")
            self.assertEqual(len(packages), 1)
            self.assertEqual(packages[0].name, "dummy")

    def test_from_empty_prefix(self):
        # Given
        with mkdtemp() as tempdir:

            # When
            repository = Repository._from_prefixes([tempdir])

            # Then
            self.assertEqual(len(list(repository.iter_packages())), 0)

    def test_delete_non_existing(self):
        # Given
        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")
        to_remove = PackageMetadata.from_egg(path)
        repository = Repository()

        # When/Then
        with self.assertRaises(NoSuchPackage):
            repository.delete_package(to_remove)

    def test_delete_simple(self):
        # Given
        eggs = ["flake8-2.0.0-2.egg", "nose-1.3.0-1.egg", "nose-1.2.1-1.egg"]
        repository = Repository()
        for egg in eggs:
            path = os.path.join(_EGGINST_COMMON_DATA, egg)
            package = RepositoryPackageMetadata.from_egg(path)
            repository.add_package(package)

        path = os.path.join(_EGGINST_COMMON_DATA, "nose-1.3.0-1.egg")
        to_remove = PackageMetadata.from_egg(path)

        # When
        repository.delete_package(to_remove)

        # Then
        assertCountEqual(self, [p.key for p in repository.iter_packages()],
                         ["flake8-2.0.0-2.egg", "nose-1.2.1-1.egg"])

    def test_find_package_from_requirement_name_only(self):
        # Given
        requirement = Requirement.from_anything("nose")

        # When
        package = self.repository.find_package_from_requirement(requirement)

        # Then
        self.assertEqual(package.full_version, "1.3.0-2")

    def test_find_package_from_requirement_name_and_version(self):
        # Given
        requirement = Requirement.from_anything("nose 1.3.0")

        # When
        package = self.repository.find_package_from_requirement(requirement)

        # Then
        self.assertEqual(package.full_version, "1.3.0-2")

        # Given
        requirement = Requirement.from_anything("nose 1.2.1")

        # When
        package = self.repository.find_package_from_requirement(requirement)

        # Then
        self.assertEqual(package.full_version, "1.2.1-1")

    def test_find_package_from_requirement_missing(self):
        # Given
        requirement_strings = ["fubar", "nose 1.3.1"]

        # When
        for requirement_string in requirement_strings:
            requirement = Requirement.from_anything(requirement_string)
            with self.assertRaises(NoSuchPackage):
                self.repository.find_package_from_requirement(requirement)

    def test_find_package_from_requirement_all(self):
        # Given
        requirement = Requirement.from_anything("nose 1.3.0-1")

        # When
        package = self.repository.find_package_from_requirement(requirement)

        # Then
        self.assertEqual(package.full_version, "1.3.0-1")

        # Given
        requirement = Requirement.from_anything("nose 1.2.1-1")

        # When
        package = self.repository.find_package_from_requirement(requirement)

        # Then
        self.assertEqual(package.full_version, "1.2.1-1")


# Unittest that used to belong to Enpkg
class TestRepositoryMisc(unittest.TestCase):
    def test_find_packages_invalid_versions(self):
        # Given
        entries = [
            dummy_installed_package_factory("numpy", "1.6.1", 1),
            dummy_installed_package_factory("numpy", "1.8k", 2),
        ]
        repository = Repository()
        for entry in entries:
            repository.add_package(entry)

        # When
        packages = repository.find_packages("numpy")

        # Then
        self.assertEqual(len(packages), 2)
        assertCountEqual(self, packages, entries)

    def test_sorted_packages_valid(self):
        # Given
        entries = [
            dummy_installed_package_factory("numpy", "1.6.1", 1),
            dummy_installed_package_factory("numpy", "1.8.0", 2),
            dummy_installed_package_factory("numpy", "1.7.1", 1),
        ]
        repository = Repository()
        for entry in entries:
            repository.add_package(entry)

        # When
        packages = repository.find_sorted_packages("numpy")

        # Then
        self.assertEqual(len(packages), 3)
        self.assertEqual([p.version for p in packages],
                         [EnpkgVersion.from_string(v)
                          for v in ("1.6.1-1", "1.7.1-1", "1.8.0-2")])

    def test_sorted_packages_invalid(self):
        # Given
        entries = [
            dummy_installed_package_factory("numpy", "1.6.1", 1),
            dummy_installed_package_factory("numpy", "1.8k", 2),
        ]
        repository = Repository()
        for entry in entries:
            repository.add_package(entry)

        # When
        packages = repository.find_sorted_packages("numpy")

        # Then
        self.assertEqual(len(packages), 2)
        assertCountEqual(self, [p.version for p in packages],
                         [EnpkgVersion.from_string(v)
                          for v in ("1.6.1-1", "1.8k-2")])



"""CacheControl import Interface.

Make it easy to import from cachecontrol without long namespaces.
"""
from .wrapper import CacheControl
from .adapter import CacheControlAdapter
from .controller import CacheController



"""
A few utilities for python requests.
"""
import base64
import logging
import os
import sqlite3

from io import FileIO

from egginst.vendor.six.moves import urllib
from egginst._compat import buffer

from enstaller.utils import uri_to_path
from enstaller.vendor import requests, sqlite_cache
from enstaller.vendor.cachecontrol.cache import BaseCache
from enstaller.vendor.cachecontrol.controller import CacheController


logger = logging.getLogger(__name__)


class FileResponse(FileIO):
    """
    A FileIO subclass that can be used as an argument to
    HTTPAdapter.build_response method from the requests library.
    """
    def __init__(self, name, mode, closefd=True):
        super(FileResponse, self).__init__(name, mode, closefd)

        self.status = 200
        self.headers = {}
        self.reason = None

    def get_all(self, name, default):
        result = self.headers.get(name)
        if not result:
            return default
        return [result]

    def getheaders(self, name):
        return self.get_all(name, [])

    def release_conn(self):
        self.close()


class LocalFileAdapter(requests.adapters.HTTPAdapter):
    """
    A requests adapter to support local file IO.

    Example
    -------

    session = requests.session()
    session.mount("file://", LocalFileAdapter())

    session.get("file:///bin/ls")
    """
    def build_response_from_file(self, request, stream):
        path = uri_to_path(request.url)
        stat_info = os.stat(path)

        response = self.build_response(request, FileResponse(path, "rb"))
        response.headers["content-length"] = str(stat_info.st_size)
        return response

    def send(self, request, stream=False, timeout=None,
             verify=True, cert=None, proxies=None):

        return self.build_response_from_file(request, stream)


class _ResponseIterator(object):
    """
    A simple iterator on top of a requests response

    It supports the `len` protocol so that packages such as click can show an
    ETA when fetching by chunk.

    Example
    -------
    >>> resp = requests.get("http://acme.com", stream=True)
    >>> for chunk in _ResponseIterator(resp):
        print len(chunk)
    """
    def __init__(self, response):
        self._response = response
        self._size = int(self._response.headers.get("content-length", 0))
        self._chunk_size = 1024

    def __iter__(self):
        self._iter = self._response.iter_content(self._chunk_size)
        return self

    def __next__(self):
        return next(self._iter)

    next = __next__

    def __len__(self):
        return int(self._size / self._chunk_size + 1)


class _NullCache(object):
    def close(self):
        pass

    def get(self, key):
        return None

    def set(self, key, value):
        pass

    def delete(self, key):
        pass


class DBCache(BaseCache):
    """
    A Sqlite-backed cache.

    Using sqlite guarantees data consistency without much overhead and
    without the need of usually brittle file locks, or external services
    (impractical in many cases)
    """
    def __init__(self, uri=":memory:", capacity=10):
        try:
            self._cache = sqlite_cache.SQLiteCache(uri, capacity,
                                                   use_separate_connection=True)
        except sqlite3.Error as e:
            logger.warn("Could not create sqlite cache: %r", e)
            self._cache = _NullCache()
        self.closed = False

    def close(self):
        if not self.closed:
            self._cache.close()
            self.closed = True

    def _encode_key(self, key):
        return base64.b64encode(key.encode("utf8")).decode("utf8")

    def _encode_value(self, value):
        data = base64.b64encode(value)
        return buffer(data)

    def _decode_value(self, encoded_value):
        return base64.b64decode(bytes(encoded_value))

    def get(self, key):
        try:
            encoded_value = self._cache.get(self._encode_key(key))
        except sqlite3.Error as e:
            logger.warn("Could not fetch data from cache: %r", e)
            return None
        else:
            if encoded_value is not None:
                try:
                    return self._decode_value(encoded_value)
                except UnicodeDecodeError as e:
                    msg = "Could not fetch data from cache: (decoding error)"
                    logger.warn(msg)
                    return None
                except Exception as e:
                    msg = "Could not fetch data from cache: {0!r}".format(e)
                    logger.warn(msg)
                    return None
            else:
                return None

    def set(self, key, value):
        try:
            self._cache.set(self._encode_key(key), self._encode_value(value))
        except sqlite3.Error as e:
            logger.warn("Could not fetch data from cache: %r", e)

    def delete(self, key):
        try:
            self._cache.delete(self._encode_key(key))
        except sqlite3.Error as e:
            logger.warn("Could not fetch data from cache: %r", e)


class QueryPathOnlyCacheController(CacheController):
    """
    A cache controller that caches entries based solely on scheme, hostname and
    path.
    """
    def cache_url(self, uri):
        url = super(QueryPathOnlyCacheController, self).cache_url(uri)
        p = urllib.parse.urlparse(url)
        return urllib.parse.urlunparse((p.scheme, p.hostname, p.path, "", "", ""))



import json
import textwrap

import mock

from egginst.vendor.six.moves import unittest

from enstaller.config import Configuration
from enstaller.plat import custom_plat
from enstaller.session import Session
from enstaller.tests.common import (mock_print,
                                    dummy_repository_package_factory,
                                    R_JSON_AUTH_RESP)
from enstaller.vendor import responses

from enstaller.tools.list_dependencies import query_platform, main


def _mock_index(entries, platform=None):
    platform = platform or custom_plat
    index = dict((entry.key, entry.s3index_data) for entry in entries)

    responses.add(responses.GET,
                  "https://api.enthought.com/eggs/{0}/index.json".
                  format(platform),
                  body=json.dumps(index), status=200,
                  content_type='application/json')


class TestListDependencies(unittest.TestCase):
    @responses.activate
    def test_simple(self):
        # Given
        config = Configuration()
        session = Session.from_configuration(config)
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 1,
                                             dependencies=["MKL 10.3"]),
        ]
        _mock_index(entries)
        r_output = textwrap.dedent("""\
            Resolving dependencies for numpy: numpy-1.8.0-1.egg
                mkl 10.3
        """)

        # When
        with mock_print() as m:
            query_platform(session, config.indices, "numpy", custom_plat)

        # Then
        self.assertMultiLineEqual(m.value, r_output)


def _mock_list_dependencies_configuration(f):
    config = Configuration()
    config.update(auth=("fake", "auth"))
    return mock.patch("enstaller.main.Configuration._from_legacy_locations",
                      return_value=config)(f)


class TestMainListDependencies(unittest.TestCase):
    def _mock_auth(self):
        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP), status=200,
                      content_type='application/json')

    @responses.activate
    @_mock_list_dependencies_configuration
    def test_another_platform(self, config):
        # Given
        self._mock_auth()
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 1,
                                             dependencies=["MKL 10.3"]),
        ]
        _mock_index(entries, "rh5-64")
        _mock_index([], "rh5-32")
        r_output = textwrap.dedent("""\
            Resolving dependencies for numpy: numpy-1.8.0-1.egg
                mkl 10.3
        """)

        # When
        with mock_print() as m:
            main(["numpy", "--platform=rh5-32"])

        # Then
        self.assertMultiLineEqual(m.value,
                                  "No egg found for requirement numpy\n")

        # When
        with mock_print() as m:
            main(["numpy", "--platform=rh5-64"])

        # Then
        self.assertMultiLineEqual(m.value, r_output)

    @responses.activate
    @_mock_list_dependencies_configuration
    def test_simple(self, config):
        # Given
        self._mock_auth()
        entries = [
            dummy_repository_package_factory("MKL", "10.3", 1),
            dummy_repository_package_factory("numpy", "1.8.0", 1,
                                             dependencies=["MKL 10.3"]),
        ]
        _mock_index(entries)
        r_output = textwrap.dedent("""\
            Resolving dependencies for numpy: numpy-1.8.0-1.egg
                mkl 10.3
        """)

        # When
        with mock_print() as m:
            main(["numpy"])

        # Then
        self.assertMultiLineEqual(m.value, r_output)



import json
import mock
import os
import shutil
import tempfile

import enstaller

from egginst.vendor.six.moves import unittest

from enstaller.auth.auth_managers import (BroodAuthenticator,
                                          LegacyCanopyAuthManager,
                                          OldRepoAuthManager)
from enstaller.config import Configuration
from enstaller.errors import EnstallerException
from enstaller.session import _PatchedRawSession, Session
from enstaller.tests.common import R_JSON_AUTH_RESP, mocked_session_factory
from enstaller.vendor import responses
from enstaller.vendor.cachecontrol.adapter import CacheControlAdapter
from enstaller.vendor.requests.adapters import HTTPAdapter


class Test_PatchedRawSession(unittest.TestCase):
    def test_mount_simple(self):
        # Given
        session = _PatchedRawSession()
        fake_adapter = mock.Mock()

        # When
        session.mount("http://", fake_adapter)

        # Then
        self.assertIs(session.adapters["http://"], fake_adapter)
        self.assertIsNot(session.adapters["https://"], fake_adapter)

    def test_umount_simple(self):
        # Given
        session = _PatchedRawSession()
        old_adapters = session.adapters.copy()
        fake_adapter = mock.Mock()

        # When
        session.mount("http://", fake_adapter)
        adapter = session.umount("http://")

        # Then
        self.assertIs(adapter, fake_adapter)
        self.assertIsNot(session.adapters["http://"], fake_adapter)
        self.assertIs(session.adapters["http://"], old_adapters["http://"])

    def test_nested_umount(self):
        # Given
        session = _PatchedRawSession()
        old_adapters = session.adapters.copy()
        fake_adapter1 = mock.Mock()
        fake_adapter2 = mock.Mock()

        # When
        session.mount("http://", fake_adapter1)
        session.mount("http://", fake_adapter2)

        # Then
        self.assertIs(session.adapters["http://"], fake_adapter2)

        # When
        adapter = session.umount("http://")

        # Then
        self.assertIs(session.adapters["http://"], fake_adapter1)
        self.assertIs(adapter, fake_adapter2)

        # When
        adapter = session.umount("http://")

        # Then
        self.assertIs(adapter, fake_adapter1)
        self.assertIs(session.adapters["http://"], old_adapters["http://"])


class TestSession(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.prefix)

    def test_etag(self):
        # Given
        config = Configuration()
        session = mocked_session_factory(config.repository_cache)

        # When
        with session.etag():
            pass

        # Then
        self.assertFalse(isinstance(session._raw.adapters["http://"],
                                    CacheControlAdapter))
        self.assertFalse(isinstance(session._raw.adapters["https://"],
                                    CacheControlAdapter))

    def _assert_use_etag_cache_controller(self, session):
        for prefix in ("http://", "https://"):
            self.assertTrue(isinstance(session._raw.adapters[prefix],
                                       CacheControlAdapter))

    def _assert_use_default_adapter(self, session):
        for prefix in ("http://", "https://"):
            self.assertTrue(isinstance(session._raw.adapters[prefix],
                                       HTTPAdapter))

    def test_nested_etag(self):
        # Given
        session = mocked_session_factory(self.prefix)

        # When/Then
        with session.etag():
            self._assert_use_etag_cache_controller(session)
            with session.etag():
                self._assert_use_etag_cache_controller(session)
            self._assert_use_etag_cache_controller(session)

            with session.etag():
                self._assert_use_etag_cache_controller(session)
                with session.etag():
                    self._assert_use_etag_cache_controller(session)
                self._assert_use_etag_cache_controller(session)
        self._assert_use_default_adapter(session)

    def test_multiple_etag(self):
        # Given
        config = Configuration()
        session = mocked_session_factory(config.repository_cache)

        # When
        with mock.patch("enstaller.session.CacheControlAdapter") as m:
            with session.etag():
                pass
            with session.etag():
                pass

        # Then
        self.assertFalse(isinstance(session._raw.adapters["http://"],
                                    CacheControlAdapter))
        self.assertFalse(isinstance(session._raw.adapters["https://"],
                                    CacheControlAdapter))
        self.assertEqual(m.call_count, 4)

    def test_max_retries(self):
        # Given
        config = Configuration()

        # When/Then
        with Session.from_configuration(config) as session:
            for prefix in ("http://", "https://"):
                self.assertEqual(session._raw.adapters[prefix].max_retries, 0)

        # When/Then
        config.update(max_retries=3)
        with Session.from_configuration(config) as session:
            for prefix in ("http://", "https://"):
                self.assertEqual(session._raw.adapters[prefix].max_retries, 3)

    def test_max_retries_with_etag(self):
        # Given
        config = Configuration()

        # When/Then
        with Session.from_configuration(config) as session:
            with session.etag():
                for prefix in ("http://", "https://"):
                    self.assertEqual(session._raw.adapters[prefix].max_retries, 0)

        # When/Then
        config.update(max_retries=3)
        with Session.from_configuration(config) as session:
            with session.etag():
                for prefix in ("http://", "https://"):
                    self.assertEqual(session._raw.adapters[prefix].max_retries, 3)

    def test_from_configuration(self):
        # Given
        config = Configuration()

        # When/Then
        with Session.from_configuration(config) as session:
            self.assertTrue(session._raw.verify)
            self.assertIsInstance(session._authenticator, LegacyCanopyAuthManager)

        # When/Then
        config = Configuration()
        config.update(verify_ssl=False)
        with Session.from_configuration(config) as session:
            self.assertFalse(session._raw.verify)

        # Given
        config = Configuration()
        config.update(verify_ssl=False, use_webservice=False)

        # When/Then
        with Session.from_configuration(config) as session:
            self.assertFalse(session._raw.verify)
            self.assertIsInstance(session._authenticator, OldRepoAuthManager)

        # Given
        config = Configuration()
        config.update(store_url="brood+http://acme.com")

        # When/Then
        with Session.from_configuration(config) as session:
            self.assertIsInstance(session._authenticator, BroodAuthenticator)

    @responses.activate
    def test_authenticated_from_configuration(self):
        # Given
        url = "http://acme.com"
        responses.add(responses.GET, url)
        responses.add(responses.GET, url + "/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP))

        config = Configuration()
        config.update(store_url=url, auth=("yoyo", "yeye"))

        # When
        with Session.authenticated_from_configuration(config) as session:
            resp = session._raw_get(url)

        # Then
        self.assertTrue(resp.status_code, 200)

    @responses.activate
    def test_authenticated_from_configuration_wo_auth(self):
        # Given
        url = "http://acme.com"
        responses.add(responses.GET, url)
        responses.add(responses.GET, url + "/accounts/user/info/",
                      body=json.dumps(R_JSON_AUTH_RESP))

        config = Configuration()
        config.update(store_url=url)

        # When/Then
        with self.assertRaises(EnstallerException):
            with Session.authenticated_from_configuration(config):
                pass

    @responses.activate
    def test_agent(self):
        # Given
        url = "http://acme.com"
        responses.add(responses.GET, url)
        config = Configuration()
        r_user_agent = "enstaller/{0}".format(enstaller.__version__)

        # When/Then
        with Session.from_configuration(config) as session:
            resp = session._raw_get(url)
            self.assertTrue(resp.request.headers["user-agent"].
                            startswith(r_user_agent))

    @responses.activate
    def test_download(self):
        # Given
        url = "http://acme.com/foo.bin"
        responses.add(responses.GET, url, body=b"some data")

        config = Configuration()

        # When
        with Session.from_configuration(config) as session:
            target = session.download(url)

        # Then
        try:
            self.assertTrue(os.path.exists(target))
            with open(target, "rb") as fp:
                self.assertEqual(fp.read(), b"some data")
        finally:
            os.unlink(target)

        self.assertEqual(target, "foo.bin")

        # When
        with Session.from_configuration(config) as session:
            target = session.download(url, "foo.baz")

        # Then
        try:
            self.assertTrue(os.path.exists(target))
            with open(target, "rb") as fp:
                self.assertEqual(fp.read(), b"some data")
        finally:
            os.unlink(target)

        self.assertEqual(target, "foo.baz")



import contextlib
import os
import shutil
import subprocess
import sys
import tempfile

SUPPORT_SYMLINK = hasattr(os, "symlink")

MACHO_DIRECTORY = os.path.join(os.path.dirname(__file__), "data", "macho")

LEGACY_PLACEHOLD_FILE_RPATH = os.path.join(MACHO_DIRECTORY,
                                           "foo_legacy_placehold_lc_rpath.dylib")
LEGACY_PLACEHOLD_FILE_LOAD_DYLIB = os.path.join(MACHO_DIRECTORY,
                                                "foo_legacy_placehold_lc_dylib.dylib")
NOLEGACY_RPATH_FILE = os.path.join(MACHO_DIRECTORY, "foo_rpath.dylib")

PYEXT_WITH_LEGACY_PLACEHOLD_DEPENDENCY = os.path.join(MACHO_DIRECTORY, "foo.so")
PYEXT_DEPENDENCY = os.path.join(MACHO_DIRECTORY, "libfoo.dylib")

FILE_TO_RPATHS = {
    NOLEGACY_RPATH_FILE: ["@loader_path/../lib"],
    LEGACY_PLACEHOLD_FILE_RPATH: ["/PLACEHOLD" * 20],
}

MACHO_ARCH_TO_FILE = {
    "x86": os.path.join(MACHO_DIRECTORY, "foo_x86"),
    "amd64": os.path.join(MACHO_DIRECTORY, "foo_amd64"),
}

PYTHON_VERSION = ".".join(str(i) for i in sys.version_info[:2])

DUMMY_EGG_WITH_INST_TARGETS = os.path.join(MACHO_DIRECTORY, "dummy_with_target_dat-1.0.0-1.egg")

_EGGINST_COMMON_DATA = os.path.join(os.path.abspath(os.path.dirname(__file__)), "data")
DUMMY_EGG_WITH_APPINST = os.path.join(_EGGINST_COMMON_DATA, "dummy_with_appinst-1.0.0-1.egg")

DUMMY_EGG = os.path.join(_EGGINST_COMMON_DATA, "dummy-1.0.1-1.egg")
DUMMY_EGG_METADATA_FILES = ("PKG-INFO", "spec/depend", "spec/summary")

DUMMY_EGG_WITH_ENTRY_POINTS = os.path.join(_EGGINST_COMMON_DATA, "dummy_with_entry_points-1.0.0-1.egg")
DUMMY_WITH_PROXY_EGG = os.path.join(_EGGINST_COMMON_DATA, "dummy_with_proxy-1.3.40-3.egg")

__st = os.stat(DUMMY_EGG)
DUMMY_EGG_MTIME = __st.st_mtime
DUMMY_EGG_SIZE = __st.st_size
DUMMY_EGG_MD5 = "1ec1f69526c55db7420b0d480c9b955e"

NOSE_1_2_1 = os.path.join(os.path.dirname(__file__), "data", "nose-1.2.1-1.egg")
NOSE_1_3_0 = os.path.join(os.path.dirname(__file__), "data", "nose-1.3.0-1.egg")

MKL_10_3 = os.path.join(os.path.dirname(__file__), "data", "MKL-10.3-1.egg")

STANDARD_EGG = os.path.join(_EGGINST_COMMON_DATA, "Jinja2-2.6-py2.7.egg")
STANDARD_EGG_METADATA_FILES = (
    "dependency_links.txt", "entry_points.txt", "not-zip-safe", "PKG-INFO",
    "requires.txt", "SOURCES.txt", "top_level.txt"
)

STANDARD_EGG_WITH_EXT = os.path.join(_EGGINST_COMMON_DATA, "PyYAML-3.11-py2.7-linux-x86_64.egg")

LEGACY_EGG_INFO_EGG = os.path.join(_EGGINST_COMMON_DATA, "flake8-2.0.0-2.egg")
LEGACY_EGG_INFO_EGG_METADATA_FILES = (
    "PKG-INFO", "requires.txt", "SOURCES.txt", "dependency_links.txt",
    "top_level.txt", "entry_points.txt"
)

ZIP_WITH_SOFTLINK = os.path.join(_EGGINST_COMMON_DATA, "zip_with_softlink.zip")

VTK_EGG_DEFERRED_SOFTLINK = os.path.join(_EGGINST_COMMON_DATA, "VTK-5.10.1-1.egg")


@contextlib.contextmanager
def mkdtemp():
    d = tempfile.mkdtemp()
    yield d
    shutil.rmtree(d)


def create_venv(prefix):
    if os.environ.get("ENSTALLER_TEST_USE_VENV", None):
        cmd = ["venv", "-s", prefix]
    else:
        cmd = ["virtualenv", "-p", sys.executable, prefix]
    subprocess.check_call(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)



import os.path
import json
import sqlite3
import tempfile

import mock

from egginst.utils import rm_rf
from egginst.vendor.six.moves import cPickle, unittest

from enstaller.requests_utils import _ResponseIterator
from enstaller.requests_utils import DBCache, FileResponse, LocalFileAdapter
from enstaller.requests_utils import _NullCache
from enstaller.utils import compute_md5
from enstaller.vendor import requests


class TestFileResponse(unittest.TestCase):
    def test_simple(self):
        # Given
        path = __file__
        with open(path, "rb") as fp:
            r_data = fp.read()

        resp = FileResponse(path, "rb")

        # When
        data = resp.read()

        # Then
        self.assertEqual(data, r_data)

    def test_getheaders(self):
        # Given
        resp = FileResponse(__file__, "rb")

        # When
        header = resp.getheaders("dummy")

        # Then
        self.assertEqual(header, [])

    def test_release_conn(self):
        # Given
        resp = FileResponse(__file__, "rb")

        # When
        resp.release_conn()

        # Then
        self.assertTrue(resp.closed)


class TestLocalFileAdapter(unittest.TestCase):
    def _create_file_session(self):
        session = requests.session()
        session.mount("file://", LocalFileAdapter())

        return session

    def test_simple(self):
        # Given
        session = self._create_file_session()
        with open(__file__, "rb") as fp:
            r_data = fp.read()

        # When
        resp = session.get("file://{0}".format(os.path.abspath(__file__)))
        data = resp.content

        # Then
        self.assertEqual(data, r_data)
        self.assertEqual(resp.headers["content-length"], str(os.stat(__file__).st_size))


class TestResponseIterator(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        rm_rf(self.prefix)

    def _create_file_session(self):
        session = requests.session()
        session.mount("file://", LocalFileAdapter())

        return session

    def test_simple(self):
        # Given
        session = self._create_file_session()
        source = __file__
        target = os.path.join(self.prefix, "output.dat")

        # When
        resp = session.get("file://{0}".format(os.path.abspath(source)))
        with open(target, "wb") as fp:
            for chunk in _ResponseIterator(resp):
                fp.write(chunk)

        # Then
        self.assertEqual(compute_md5(target), compute_md5(source))

    def test_without_content_length(self):
        # Given
        session = self._create_file_session()
        source = __file__
        target = os.path.join(self.prefix, "output.dat")

        # When
        resp = session.get("file://{0}".format(os.path.abspath(source)))
        resp.headers.pop("content-length")
        with open(target, "wb") as fp:
            for chunk in _ResponseIterator(resp):
                fp.write(chunk)

        # Then
        self.assertEqual(compute_md5(target), compute_md5(source))


class TestDBCache(unittest.TestCase):
    def setUp(self):
        self.prefix = tempfile.mkdtemp()

    def tearDown(self):
        rm_rf(self.prefix)

    def test_cant_write_db(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        with open(uri, "wb") as fp:
            fp.write(b"")
        os.chmod(uri, 0o000)

        cache = DBCache(uri)

        # When/Then
        # Ensure we don't get an error when using the null cache
        try:
            cache.get("foo")
        finally:
            cache.close()

        # Then
        self.assertIsInstance(cache._cache, _NullCache)

    def test_simple(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        cache = DBCache(uri)
        r_value = json.dumps({"bar": "fubar"}).encode("utf8")

        # When
        cache.set("foo", r_value)
        value = cache.get("foo")

        # Then
        self.assertEqual(value, r_value)

        # When
        cache.delete("foo")
        value = cache.get("foo")

        # Then
        self.assertIsNone(value)

        # When/Then
        # Ensure we don't raise an exception when deleting twice
        cache.delete("foo")

    def test_simple_pickle(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        cache = DBCache(uri)
        r_value = cPickle.dumps({"bar": "fubar"}, cPickle.HIGHEST_PROTOCOL)

        # When
        cache.set("foo", r_value)
        value = cache.get("foo")

        # Then
        self.assertEqual(value, r_value)

        # When
        cache.delete("foo")
        value = cache.get("foo")

        # Then
        self.assertIsNone(value)

        # When/Then
        # Ensure we don't raise an exception when deleting twice
        cache.delete("foo")

    def test_simple_cannot_create_db(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        with open(uri, "wb") as fp:
            fp.write(b"")
        os.chmod(uri, 0o500)

        cache = DBCache(uri)
        r_value = json.dumps({"bar": "fubar"}).encode("utf8")

        # When
        cache.set("foo", r_value)
        value = cache.get("foo")

        # Then
        self.assertIsNone(value)

    def _create_invalid_table(self, uri):
        cx = sqlite3.Connection(uri)
        cx.execute("""\
CREATE TABLE queue
(
    id INTEGER PRIMARY KEY AUTOINCREMENT
);""")

    def test_simple_invalid_db(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        self._create_invalid_table(uri)

        cache = DBCache(uri)

        # When
        cache.set("foo", b"bar")
        value = cache.get("foo")

        # Then
        self.assertIsNone(value)

        # When/Then
        cache.delete("foo")

    def test_invalid_data(self):
        # Given
        uri = os.path.join(self.prefix, "foo.db")
        with open(uri, "wb") as fp:
            fp.write(b"")

        cache = DBCache(uri)
        r_value = b"bar"

        # When
        cache.set("foo", r_value)

        with mock.patch("enstaller.requests_utils.base64.b64decode",
                        side_effect=TypeError):
            value = cache.get("foo")

        # Then
        self.assertIsNone(value)



from __future__ import absolute_import

import contextlib
import json
import os.path
import re
import shutil
import sys
import tempfile
import textwrap

if sys.version_info[:2] < (2, 7):
    import unittest2 as unittest
else:
    import unittest

import mock

from enstaller.vendor import responses

from enstaller.auth import UserPasswordAuth
from enstaller.main import main, main_noexc
from enstaller.config import (_set_keyring_password,
                              Configuration, write_default_config)

from enstaller.tests.common import (fake_keyring, mock_print, mock_input_auth,
                                    mock_raw_input)
from enstaller.vendor import requests
from enstaller.auth.tests.test_auth import R_JSON_AUTH_RESP

from .common import (
    empty_index, mock_install_req, use_given_config_context,
    without_any_configuration)

FAKE_USER = "nono"
FAKE_PASSWORD = "le petit robot"
FAKE_CREDS = UserPasswordAuth(FAKE_USER, FAKE_PASSWORD)._encoded_auth

class TestAuth(unittest.TestCase):
    @contextlib.contextmanager
    def assertSuccessfulExit(self):
        with self.assertRaises(SystemExit) as e:
            yield e
        self.assertEqual(e.exception.code, 0)

    def setUp(self):
        self.d = tempfile.mkdtemp()
        self.config = os.path.join(self.d, ".enstaller4rc")
        write_default_config(self.config)

    def tearDown(self):
        shutil.rmtree(self.d)

    def _run_main_with_dummy_req(self):
        with use_given_config_context(self.config):
            with self.assertSuccessfulExit():
                main_noexc(["dummy_requirement"])

    def _dummy_ssl_error_callback(self, request):
        msg = "Dummy SSL error"
        raise requests.exceptions.SSLError(msg, request=request)

    def test_auth_requested_without_config(self):
        """
        Ensure we ask for authentication if no .enstaller4rc is found.
        """
        with use_given_config_context(self.config):
            with mock_print() as m:
                with mock.patch("enstaller.main.configure_authentication_or_exit",
                                lambda *a: sys.exit(-1)):
                    with self.assertRaises(SystemExit):
                        main_noexc([])

        self.assertMultiLineEqual(m.value, "")

    @without_any_configuration
    def test_userpass_without_config(self):
        """
        Ensure we don't crash when empty information is input in --userpass
        prompt (no .enstaller4rc found).
        """
        from enstaller.main import main
        with use_given_config_context(self.config):
            with mock_input_auth("", "") as m:
                with self.assertRaises(SystemExit):
                    main_noexc(["--userpass"])

        self.assertEqual(m.call_count, 3)

    @responses.activate
    def test_userpass_with_config(self):
        """
        Ensure enpkg --userpass doesn't crash when creds are invalid
        """
        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      status=401,
                      content_type='application/json')
        r_output = textwrap.dedent("""\
        Could not authenticate as 'nono'
        Please check your credentials/configuration and try again
        (original error is: '401 Client Error: None').


        No modification was written.
        """)

        with use_given_config_context(self.config):
            with mock_print() as m:
                with mock_input_auth("nono", "robot"):
                    with self.assertRaises(SystemExit):
                        main_noexc(["--userpass"])

        self.assertMultiLineEqual(m.value, r_output)

    @mock_install_req
    @empty_index
    @responses.activate
    def test_enpkg_req_with_valid_auth(self):
        """
        Ensure 'enpkg req' authenticate as expected
        """
        with open(self.config, "w") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))

        json_data = {
            "is_authenticated": True,
            "first_name": "nono",
            "last_name": "le petit robot",
            "has_subscription": True,
            "subscription_level": "free"
        }
        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      body=json.dumps(json_data),
                      status=200,
                      content_type='application/json')

        with use_given_config_context(self.config):
            main(["nono"])

    @responses.activate
    def test_enpkg_req_with_invalid_auth(self):
        """
        Ensure 'enpkg req' doesn't crash when creds are invalid
        """
        self.maxDiff = None
        responses.add(responses.GET,
                      "https://api.enthought.com/accounts/user/info/",
                      status=401,
                      content_type='application/json')
        r_output = textwrap.dedent("""\
            Could not authenticate as 'nono'
            Please check your credentials/configuration and try again
            (original error is: '401 Client Error: None').


            You can change your authentication details with 'enpkg --userpass'.
        """)

        with open(self.config, "w") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))

        with mock_print() as m:
            with use_given_config_context(self.config):
                with self.assertRaises(SystemExit):
                    main_noexc(["nono"])

        self.assertMultiLineEqual(m.value, r_output)

    @empty_index
    @mock_install_req
    @fake_keyring
    @responses.activate
    def test_no_keyring_to_no_keyring_conversion(self):
        """
        Ensure the config file is not converted when configured not to use
        keyring.
        """
        # Given
        with open(self.config, "w") as fp:
            fp.write("EPD_auth = '{0}'".format(FAKE_CREDS))
        config = Configuration.from_file(self.config)
        responses.add(responses.GET, config.api_url, status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))

        # When
        self._run_main_with_dummy_req()

        # Then
        with open(self.config) as fp:
            self.assertMultiLineEqual(fp.read(), "EPD_auth = '{0}'".format(FAKE_CREDS))

    @empty_index
    @mock_install_req
    @fake_keyring
    @responses.activate
    def test_keyring_to_no_keyring_conversion(self):
        """
        Ensure the config file is automatically converted to use keyring.
        """
        # Given
        _set_keyring_password(FAKE_USER, FAKE_PASSWORD)
        with open(self.config, "w") as fp:
            fp.write("EPD_username = '{0}'".format(FAKE_USER))
        config = Configuration.from_file(self.config)
        responses.add(responses.GET, config.api_url, status=200,
                      body=json.dumps(R_JSON_AUTH_RESP))

        # When
        self._run_main_with_dummy_req()

        # Then
        with open(self.config) as fp:
            self.assertMultiLineEqual(fp.read(), "EPD_auth = '{0}'".format(FAKE_CREDS))


    @mock_install_req
    @fake_keyring
    @responses.activate
    def test_401_index_handling(self):
        self.maxDiff = None

        # Given
        repo = "http://acme.com/repo/ets/"
        config = Configuration()
        config.update(use_webservice=False, indexed_repositories=[repo])
        config.update(auth=("nono", "le petit robot"))
        config.write(self.config)

        url = repo + "index.json"
        responses.add(responses.HEAD, url, status=401)

        error_message = textwrap.dedent("""\
            Could not authenticate as 'nono'
            Please check your credentials/configuration and try again
            (original error is: '401 Client Error: None').


            You can change your authentication details with 'enpkg --userpass'.
        """)

        # When
        with use_given_config_context(self.config):
            with mock_print() as m:
                with self.assertRaises(SystemExit):
                    main(["dummy_requirement"])

        # Then
        self.assertMultiLineEqual(m.value, error_message)

    @mock_install_req
    @fake_keyring
    @responses.activate
    def test_invalid_certificate_use_webservice(self):
        self.maxDiff = None

        # Given
        url = "http://acme.com"
        config = Configuration()
        config.update(store_url=url)
        config.update(auth=("nono", "le petit robot"))
        config.write(self.config)

        def callback(request):
            msg = "Dummy SSL error"
            raise requests.exceptions.SSLError(msg, request)
        responses.add_callback(responses.GET, re.compile(url + "/*"),
                               callback)

        error_message = textwrap.dedent("""\
            SSL error: [Errno Dummy SSL error] <PreparedRequest [GET]>
            To connect to 'acme.com' insecurely, add the `-k` flag to enpkg command
        """)

        # When
        with use_given_config_context(self.config):
            with mock_print() as m:
                with self.assertRaises(SystemExit):
                    main(["dummy_requirement"])

        # Then
        self.assertMultiLineEqual(m.value, error_message)

    @mock_install_req
    @fake_keyring
    @responses.activate
    def test_invalid_certificate_dont_webservice(self):
        self.maxDiff = None

        # Given
        url = "http://acme.com"
        config = Configuration()
        config.update(use_webservice=False,
                      indexed_repositories=["http://acme.com/repo/"])
        config.update(auth=("nono", "le petit robot"))
        config.write(self.config)

        def callback(request):
            msg = "Dummy SSL error"
            raise requests.exceptions.SSLError(msg, request=request)
        responses.add_callback(responses.HEAD, re.compile(url + "/*"),
                               callback)

        error_message = textwrap.dedent("""\
            SSL error: Dummy SSL error
            To connect to 'acme.com' insecurely, add the `-k` flag to enpkg command
        """)

        # When
        with use_given_config_context(self.config):
            with mock_print() as m:
                with self.assertRaises(SystemExit):
                    main(["dummy_requirement"])

        # Then
        self.assertMultiLineEqual(m.value, error_message)

    @fake_keyring
    @responses.activate
    def test_invalid_certificate_config(self):
        self.maxDiff = None

        # Given
        url = "http://acme.com"
        config = Configuration()
        config.update(store_url=url)
        config.update(auth=("nono", "le petit robot"))
        config.write(self.config)

        responses.add_callback(responses.GET, re.compile(url + "/*"),
                               self._dummy_ssl_error_callback)

        error_message = "To connect to 'acme.com' insecurely, add the `-k` " \
                        "flag to enpkg command"

        # When
        with use_given_config_context(self.config):
            with mock_print() as m:
                with self.assertRaises(SystemExit):
                    main(["--config"])

        # Then
        last_line = m.value.splitlines()[-1]
        self.assertMultiLineEqual(last_line, error_message)



#
# (C) Copyright 2014 Enthought, Inc., Austin, TX
# All right reserved.
#
# This file is open source software distributed according to the terms in
# LICENSE.txt
#

__all__ = [
    "CredWrite", "CredRead", "CRED_TYPE_GENERIC", "CRED_PERSIST_ENTERPRISE"]

import ctypes

from ._common import _PyBytes_FromStringAndSize, _GetACP
from .compat import is_unicode, unicode
from . import _win32cred

CRED_TYPE_GENERIC = 0x1
CRED_PERSIST_ENTERPRISE = 0x3


def CredWrite(credential, flag):
    """
    Mimic pywin32 win32cred.CredWrite.

    Parameters
    ----------
    credential: dict
        Dict of parameters to be passed to win32 API CredWrite
    flag: int

    Returns
    -------
    credentials: dict
        A dictionary containing the following:

            - Type: the type of credential (see MSDN)
            - TargetName: the target to use (string)
            - Persist: see MSDN
            - UserName: the retrieved username
            - CredentialBlob: the password (as a *string*, not an encoded
              binary stream - this function takes care of the encoding).
            - Comment: a string
    """

    unsupported = set(credential.keys()) - _win32cred.SUPPORTED_CREDKEYS
    if len(unsupported):
        raise ValueError("Unsupported keys: {0}".format(unsupported))
    if flag != 0:
        raise ValueError("flag != 0 not yet supported")

    c_creds = _win32cred.CREDENTIAL()
    c_pcreds = ctypes.pointer(c_creds)

    ctypes.memset(c_pcreds, flag, ctypes.sizeof(c_creds))

    for key in _win32cred.SUPPORTED_CREDKEYS:
        if key in credential:
            if key != 'CredentialBlob':
                setattr(c_creds, key, credential[key])
            else:
                blob = _make_blob(credential['CredentialBlob'])
                blob_data = ctypes.create_unicode_buffer(blob)
                # Create_unicode_buffer adds a NULL at the end of the string
                # we do not want that.
                c_creds.CredentialBlobSize = \
                    ctypes.sizeof(blob_data) - ctypes.sizeof(ctypes.c_wchar)
                c_creds.CredentialBlob = ctypes.cast(
                    blob_data, _win32cred.LPBYTE)

    _win32cred._CredWrite(c_pcreds, 0)


def CredRead(TargetName, Type):
    """
    Mimic pywin32 win32cred.CreadRead.

    Parameters
    ----------
    TargetName: str-like
        The target name to fetch from the keyring.

    Returns
    -------
    credentials: dict
        A dictionary containing the following:

            - UserName: the retrieved username
            - CredentialBlob: the password (as an utf-16 encoded 'string')

        None if the target name was not found.
    """
    if not Type == CRED_TYPE_GENERIC:
        raise ValueError("Type != CRED_TYPE_GENERIC not yet supported")

    flag = 0
    c_pcreds = _win32cred.PCREDENTIAL()

    _win32cred._CredRead(TargetName, Type, flag, ctypes.byref(c_pcreds))
    try:
        c_creds = c_pcreds.contents
        credential = {}
        for key in _win32cred.SUPPORTED_CREDKEYS:
            if key != 'CredentialBlob':
                credential[key] = getattr(c_creds, key)
            else:
                blob = _PyBytes_FromStringAndSize(
                    c_creds.CredentialBlob, c_creds.CredentialBlobSize)
                credential['CredentialBlob'] = blob
        return credential
    finally:
        _win32cred.advapi.CredFree(c_pcreds)


def CredDelete(TargetName, Type):
    """
    Remove the given target name from the stored credentials.

    Mimic pywin32 win32cred.CreadDelete.

    Parameters
    ----------
    TargetName: str-like
        The target name to fetch from the keyring.
    """
    if not Type == CRED_TYPE_GENERIC:
        raise ValueError("Type != CRED_TYPE_GENERIC not yet supported.")
    _win32cred._CredDelete(TargetName, Type, 0)


def _make_blob(password):
    """ Convert the input string password into a unicode blob as required for
    Credentials.

    """
    if is_unicode(password):
        return password
    else:
        code_page = _GetACP()
        return unicode(password, encoding=str(code_page), errors='strict')



import collections
import contextlib
import os.path
import threading

from egginst.utils import atomic_file, ensure_dir
from egginst.vendor.six.moves import urllib

from enstaller import __version__
from enstaller.auth.auth_managers import (BroodAuthenticator,
                                          LegacyCanopyAuthManager,
                                          OldRepoAuthManager)
from enstaller.config import STORE_KIND_BROOD
from enstaller.errors import EnstallerException
from enstaller.vendor import requests
from enstaller.vendor.cachecontrol.adapter import CacheControlAdapter

from enstaller.requests_utils import (DBCache, LocalFileAdapter,
                                      QueryPathOnlyCacheController)

from enstaller.auth import UserPasswordAuth


class _PatchedRawSession(requests.Session):
    """ Like requests.Session, but supporting (nested) umounting of
    adapters.
    """
    def __init__(self, *a, **kw):
        self._adapters_stack = collections.defaultdict(list)
        super(_PatchedRawSession, self).__init__(*a, **kw)

    def mount(self, prefix, adapter):
        self._adapters_stack[prefix].append(adapter)
        super(_PatchedRawSession, self).mount(prefix, adapter)

    def umount(self, prefix):
        if prefix in self._adapters_stack \
                and len(self._adapters_stack[prefix]) > 0:
            self.adapters.pop(prefix)
            current_adapter = self._adapters_stack[prefix].pop()

            if len(self._adapters_stack[prefix]) > 0:
                previous_adapter = self._adapters_stack[prefix][-1]
                self.adapters[prefix] = previous_adapter

            return current_adapter
        else:
            msg = "no adapter registered for prefix {0!r}".format(prefix)
            raise ValueError(msg)


class Session(object):
    """ Simple class to handle http session management

    It also ensures connection settings such as proxy, SSL CA certification,
    etc... are handled consistently).

    Parameters
    ----------
    authenticator : IAuthManager
        An authenticator instance
    cache_directory : str
        A writeable directory to cache data and eggs.
    proxies : dict
        A proxy dict as expected by requests (and provided by
        Configuration.proxy_dict).
    verify : bool
        If True, SSL CA are verified (default).
    """
    def __init__(self, authenticator, cache_directory, proxies=None,
                 verify=True, max_retries=0):
        self.proxies = proxies
        self.verify = verify
        self.cache_directory = cache_directory
        self.max_retries = max_retries

        self._authenticator = authenticator
        self._raw = _PatchedRawSession()
        if proxies is not None:
            self._raw.proxies = proxies
        self._raw.verify = verify

        self._raw.mount("file://", LocalFileAdapter())

        adapter = requests.adapters.HTTPAdapter(max_retries=self.max_retries)
        for prefix in ("http://", "https://"):
            self._raw.mount(prefix, adapter)

        user_agent = "enstaller/{0} {1}".format(__version__,
                                                self._raw.headers["user-agent"])
        self._raw.headers["user-agent"] = user_agent

        self._in_etag_context = 0
        self._etag_context_lock = threading.RLock()

    @classmethod
    def authenticated_from_configuration(cls, configuration):
        """ Create a new authenticated session from a configuration.

        Parameters
        ----------
        configuration : Configuration
            The configuration to use.
        """
        if configuration.auth is None:
            msg = "No auth configured for the given configuration."
            raise EnstallerException(msg)
        session = cls.from_configuration(configuration)
        session.authenticate(configuration.auth)
        return session

    @classmethod
    def from_configuration(cls, configuration):
        """ Create a new session from a configuration.

        Parameters
        ----------
        configuration : Configuration
            The configuration to use.
        """
        if configuration.store_kind == STORE_KIND_BROOD:
            klass = BroodAuthenticator
        elif configuration.use_webservice:
            klass = LegacyCanopyAuthManager
        else:
            klass = OldRepoAuthManager
        authenticator = klass.from_configuration(configuration)
        return cls(authenticator, configuration.repository_cache,
                   configuration.proxy_dict, verify=configuration.verify_ssl,
                   max_retries=configuration.max_retries)

    def close(self):
        self._raw.close()

    def __enter__(self):
        return self

    def __exit__(self, *a, **kw):
        self.close()

    @contextlib.contextmanager
    def etag(self):
        with self._etag_context_lock:
            self._etag_setup()
        try:
            yield
        finally:
            with self._etag_context_lock:
                self._etag_tear()

    def _etag_setup(self):
        if self._in_etag_context == 0:
            uri = os.path.join(self.cache_directory, "index_cache", "index.db")
            ensure_dir(uri)
            cache = DBCache(uri)

            for prefix in ("http://", "https://"):
                adapter = CacheControlAdapter(
                    cache, controller_class=QueryPathOnlyCacheController,
                    max_retries=self.max_retries)
                self._raw.mount(prefix, adapter)

        self._in_etag_context += 1

    def _etag_tear(self):
        if self._in_etag_context == 1:
            for prefix in ("https://", "http://"):
                # XXX: This close is ugly, but I am not sure how one can link a cache
                # controller to a http adapter in cachecontrol. See issue #42 on
                # ionrock/cachecontrol @ github.
                adapter = self._raw.umount(prefix)
                adapter.cache.close()
        self._in_etag_context -= 1

    def authenticate(self, auth):
        if isinstance(auth, tuple) and len(auth) == 2:
            auth = UserPasswordAuth(auth[0], auth[1])

        self._authenticator.authenticate(self, auth.request_adapter)
        self._raw.auth = self._authenticator._auth

    def download(self, url, target=None):
        """ Safely download the content at the give url.

        Safely here is understood as not leaving stalled content if the
        download fails or is canceled. It uses streaming as so not to use too
        much memory.
        """
        resp = self._raw.get(url, stream=True)
        resp.raise_for_status()

        if target is None:
            target = os.path.basename(urllib.parse.urlparse(url).path)

        with atomic_file(target) as fp:
            for chunk in resp.iter_content(1024):
                fp.write(chunk)

        return target

    def fetch(self, url):
        resp = self._raw.get(url, stream=True)
        resp.raise_for_status()
        return resp

    def get(self, url, stream=False):
        return self._raw.get(url, stream=stream)

    def head(self, url):
        return self._raw.head(url)

    def _raw_get(self, url, auth=None):
        return self._raw.get(url, auth=auth)

    def _raw_head(self, url, auth=None):
        return self._raw.head(url, auth=auth)

    def _raw_post(self, url, auth=None):
        return self._raw.post(url, auth=auth)



# Author: Ilan Schnell <ischnell@enthought.com>
"""\
The enstaller package is a tool for managing egginst-based installs.
Its primary command-line interface program is enpkg, which processes user
commands and in turn invokes egginst to do the actual installations.
enpkg can access eggs from both local and HTTP repositories.
"""
from __future__ import print_function

import argparse
import logging
import ntpath
import os
import posixpath
import re
import site
import sys
import textwrap
import warnings

from argparse import ArgumentParser
from os.path import isfile

from egginst.main import EGG_INFO
from egginst.progress import console_progress_manager_factory
from egginst.vendor.six.moves import http_client

import enstaller

from enstaller.auth import UserPasswordAuth
from enstaller.errors import (EnpkgError, EnstallerException,
                              InvalidPythonPathConfiguration,
                              InvalidConfiguration,
                              EXIT_ABORTED)
from enstaller.config import (ENSTALLER4RC_FILENAME, HOME_ENSTALLER4RC,
                              Configuration, add_url,
                              configuration_read_search_order,
                              convert_auth_if_required,
                              print_config, write_default_config)
from enstaller.session import Session
from enstaller.errors import AuthFailedError, NoSuchPackage
from enstaller.enpkg import Enpkg, ProgressBarContext
from enstaller.repository import InstalledPackageMetadata, Repository
from enstaller.solver import Request, Requirement
from enstaller.utils import abs_expanduser, input_auth, prompt_yes_no
from enstaller.vendor import requests
from enstaller.versions.enpkg import EnpkgVersion

from enstaller.cli.commands import (env_option, freeze, imports_option,
                                    info_option, install_from_requirements,
                                    list_option, print_history, revert, search,
                                    update_all, whats_new)
from enstaller.cli.utils import (exit_if_root_on_non_owned,
                                 humanize_ssl_error_and_die, install_req,
                                 repository_factory)

from enstaller._update_support import inplace_update

logger = logging.getLogger(__name__)

PLEASE_AUTH_MESSAGE = ("No authentication configured, required to continue.\n"
                       "To login, type 'enpkg --userpass'.")


def epd_install_confirm(force_yes=False):
    msg = textwrap.dedent("""\
        Warning: 'enpkg epd' will downgrade any packages that are currently")
        at a higher version than in the specified EPD release.
        Usually it is preferable to update all installed packages with:
            enpkg --update-all""")
    print(msg)
    return prompt_yes_no("Are you sure that you wish to proceed? (y/[n]) ",
                         force_yes)


def _get_enstaller_comparable_version(prefix, package_name):
    runtime_version = \
        EnpkgVersion.from_upstream_and_build(enstaller.__version__, 1)

    egg_info_dir = os.path.join(prefix, EGG_INFO, package_name)
    try:
        # Installed as an egg in sys.prefix
        enstaller_package = \
            InstalledPackageMetadata.from_meta_dir(egg_info_dir)
        if str(enstaller_package.version.upstream) != enstaller.__version__:
            # It is both installed as an egg and from sources/develop mode, the
            # latter takes precedence
            current_comparable_version = runtime_version
        else:
            current_comparable_version = enstaller_package.version
    except EnstallerException:
        # Installed from sources, no build number
        current_comparable_version = runtime_version

    return current_comparable_version


def update_enstaller(session, repository, opts):
    """
    Check if Enstaller is up to date, and if not, ask the user if he
    wants to update.  Return boolean indicating whether enstaller was
    updated.
    """
    updated = False
    if not enstaller.__is_released__:
        return updated

    package_name = "enstaller"

    current_comparable_version = \
        _get_enstaller_comparable_version(sys.prefix, package_name)

    try:
        latest = repository.find_latest_package(package_name)
    except NoSuchPackage:
        updated = False
    else:
        if latest.version > current_comparable_version:
            if prompt_yes_no("Enstaller is out of date.  Update? ([y]/n) ",
                             opts.yes):
                inplace_update(session, repository, latest)
                updated = True

    return updated


def get_package_path(prefix):
    """Return site-packages path for the given repo prefix.

    Note: on windows the path is lowercased and returned.
    """
    if sys.platform == 'win32':
        return ntpath.join(prefix, 'Lib', 'site-packages').lower()
    else:
        postfix = 'lib/python{0}.{1}/site-packages'.format(*sys.version_info)
        return posixpath.join(prefix, postfix)


def check_prefixes(prefixes):
    """
    Check that package prefixes lead to site-packages that are on the python
    path and that the order of the prefixes matches the python path.
    """
    index_order = []
    if sys.platform == 'win32':
        sys_path = [x.lower() for x in sys.path]
    else:
        sys_path = sys.path
    for prefix in prefixes:
        path = get_package_path(prefix)
        try:
            index_order.append(sys_path.index(path))
        except ValueError:
            msg = "Expected to find %s in PYTHONPATH" % (path, )
            raise InvalidPythonPathConfiguration(msg)
    else:
        if not index_order == sorted(index_order):
            msg = "Order of path prefixes doesn't match PYTHONPATH"
            raise InvalidPythonPathConfiguration(msg)


def needs_to_downgrade_enstaller(reqs):
    """
    Returns True if the running enstaller would be downgraded by satisfying the
    list of requirements.
    """
    for req in reqs:
        if req.name == "enstaller" and req.version is not None:
            return True
    return False


def _get_config_candidate():
    """ Return a full path if one candidate is found, None if no config file
    was found.
    """
    paths = [os.path.join(d, ENSTALLER4RC_FILENAME) for d in
             configuration_read_search_order()]
    for path in paths:
        if isfile(path):
            return path


def _invalid_authentication_message(auth, exc):
    assert isinstance(exc, AuthFailedError)

    if exc.original_exception is None:
        original_error = str(exc)
    else:
        original_error = str(exc.original_exception)

    header = auth.cant_login_message
    msg = textwrap.dedent("""\
        {0}
        Please check your credentials/configuration and try again
        (original error is: {1!r}).
        """.format(header, str(original_error)))
    return msg


def ensure_authenticated_config(config, config_filename, session,
                                use_new_format=False):
    try:
        session.authenticate(config.auth)
    except requests.exceptions.SSLError as e:
        humanize_ssl_error_and_die(e, config.store_url)
    except AuthFailedError as exc:
        msg = _invalid_authentication_message(config.auth, exc)
        print(msg)
        print("\nYou can change your authentication details with "
              "'enpkg --userpass'.")
        sys.exit(-1)
    else:
        if not use_new_format:
            convert_auth_if_required(config_filename)


def configure_authentication_or_exit(config, config_filename,
                                     session):
    n_trials = 3
    for i in range(n_trials):
        username, password = input_auth()
        if username:
            break
        else:
            msg = "Please enter a non empty username ({0} trial(s) left, " \
                  "Ctrl+C to exit)". format(n_trials - i - 1)
            print(msg)
    else:
        print("No valid username entered (no modification was written).")
        sys.exit(-1)

    auth = UserPasswordAuth(username, password)

    try:
        config._checked_change_auth(auth, session, config_filename)
    except AuthFailedError as exc:
        msg = _invalid_authentication_message(auth, exc)
        print(msg)
        print("\nNo modification was written.")
        sys.exit(-1)


def setup_proxy_or_die(config, proxy):
    if proxy:
        try:
            config.update(proxy=proxy)
        except InvalidConfiguration as e:
            print("Error: invalid proxy setting {0!r}".format(e))
            sys.exit(1)


def dispatch_commands_without_enpkg(args, config, config_filename, prefixes,
                                    prefix, pat, session):
    """
    Returns True if a command has been executed.
    """
    if args.env:                                  # --env
        env_option(prefixes)
        return True

    if args.log:                                  # --log
        print_history(prefix)
        return True

    if args.freeze:
        freeze(prefixes)
        return True

    if args.list:                                 # --list
        list_option(prefixes, pat)
        return True

    if args.config:                               # --config
        print_config(config, session)
        return True

    if args.add_url:                              # --add-url
        add_url(config_filename, config, args.add_url)
        return True

    if args.userpass:                             # --userpass
        configure_authentication_or_exit(config, config_filename,
                                         session)
        return True


def dispatch_commands_with_enpkg(args, enpkg, config, prefix, session, parser,
                                 pat):
    if args.dry_run:
        def print_actions(actions):
            for item in actions:
                print('%-8s %s' % item)
        enpkg.execute = print_actions

    if args.imports:                              # --imports
        repository = Repository._from_prefixes(enpkg.prefixes)
        imports_option(repository)
        return

    if args.revert:                               # --revert
        revert(enpkg, args.revert)
        return

    # Try to auto-update enstaller
    if config.autoupdate and not args.no_autoupdate:
        if update_enstaller(session, enpkg._remote_repository, args):
            print("Enstaller has been updated.\n"
                  "Please re-run your previous command.")
            return

    if args.search:                               # --search
        search(enpkg._remote_repository, enpkg._installed_repository,
               config, session, pat)
        return

    if args.info:                                 # --info
        if len(args.cnames) != 1:
            parser.error("Option requires one argument (name of package)")
        info_option(enpkg._remote_repository, enpkg._installed_repository,
                    args.cnames[0])
        return

    if args.whats_new:                            # --whats-new
        whats_new(enpkg._remote_repository, enpkg._installed_repository)
        return

    if args.update_all:                           # --update-all
        update_all(enpkg, config, args)
        return

    if args.requirements:
        install_from_requirements(enpkg, config, args)
        return

    if len(args.cnames) == 0 and not args.remove_enstaller:
        parser.error("Requirement(s) missing")
    elif len(args.cnames) == 2:
        pat = re.compile(r'\d+\.\d+')
        if pat.match(args.cnames[1]):
            args.cnames = ['-'.join(args.cnames)]

    reqs = _compute_reqs(args.cnames)

    # This code assumes we have already upgraded enstaller if needed
    if needs_to_downgrade_enstaller(reqs):
        msg = "Enstaller in requirement list: enstaller will be downgraded !"
        warnings.warn(msg)
    else:
        logger.debug("Enstaller is up to date, not updating")
        reqs = [req for req in reqs if req.name != "enstaller"]

    logger.info("Requirements:")
    for req in reqs:
        logger.info('    %r', req)

    logger.info("prefix: %r", prefix)

    REMOVE_ENSTALLER_WARNING = ("Removing enstaller package will break enpkg "
                                "and is not recommended.")
    if args.remove:
        if any(req.name == 'enstaller' for req in reqs):
            print(REMOVE_ENSTALLER_WARNING)
            print("If you are sure you wish to remove enstaller, use:")
            print("    enpkg --remove-enstaller")
            return

    if args.remove_enstaller:
        print(REMOVE_ENSTALLER_WARNING)
        if prompt_yes_no("Really remove enstaller? (y/[n]) ", args.yes):
            args.remove = True
            reqs = [Requirement('enstaller')]

    if any(req.name == 'epd' for req in reqs):
        if args.remove:
            parser.error("Can't remove 'epd'")
        elif len(reqs) > 1:
            parser.error("Can't combine 'enpkg epd' with other packages.")
        elif not epd_install_confirm(args.yes):
            return

    if args.remove:
        for req in reqs:
            solver = enpkg._solver_factory()
            try:
                request = Request()
                request.remove(req)
                enpkg.execute(solver.resolve(request))
            except EnpkgError as e:
                print(str(e))
    else:
        for req in reqs:
            install_req(enpkg, config, req, args)


def _user_base():
    return getattr(site, "USER_BASE", abs_expanduser('~/.local'))


def _create_parser():
    p = ArgumentParser(description=__doc__)
    p.add_argument('cnames', metavar='NAME', nargs='*',
                   help='package(s) to work on')
    p.add_argument("--add-url", metavar='URL',
                   help="add a repository URL to the configuration file")
    p.add_argument("--insecure", "-k", action="store_true",
                   default=argparse.SUPPRESS,
                   help="Disable SSL cert verification")
    p.add_argument("--config", action="store_true",
                   help="display the configuration and exit")
    p.add_argument("-c", "--config-path",
                   help="Alternative configuration location (must be in the "
                        "new YAML-based format).")
    p.add_argument('-f', "--force", action="store_true",
                   help="force install the main package "
                        "(not its dependencies, see --forceall)")
    p.add_argument("--forceall", action="store_true",
                   help="force install of all packages "
                        "(i.e. including dependencies)")
    p.add_argument("--freeze", help=argparse.SUPPRESS, action="store_true")
    p.add_argument("--imports", action="store_true",
                   help="show which packages can be imported")
    p.add_argument('-i', "--info", action="store_true",
                   help="show information about a package")
    p.add_argument("--log", action="store_true", help="print revision log")
    p.add_argument('-l', "--list", action="store_true",
                   help="list the packages currently installed on the system")
    p.add_argument("--max-retries", type=int,
                   default=argparse.SUPPRESS,
                   help="Maximum number of retries for a checksum mismatch or "
                        "a connection error.")
    p.add_argument('-n', "--dry-run", action="store_true",
                   help="show what would have been downloaded/removed/installed")
    p.add_argument('-N', "--no-deps", action="store_true",
                   help="neither download nor install dependencies")
    p.add_argument("--env", action="store_true",
                   help="based on the configuration, display how to set "
                        "environment variables")
    p.add_argument("--prefix", metavar='PATH',
                   help="install prefix (disregarding any settings in "
                        "the config file)")
    p.add_argument("--proxy", metavar='PROXYSTR',
                   help="use a proxy for downloads."
                        " <proxy protocol>://[<proxy username>"
                        "[:<proxy password>@]]<proxy server>:<proxy port>")
    p.add_argument("--remove", action="store_true", help="remove a package")
    p.add_argument("--remove-enstaller", action="store_true",
                   help="remove enstaller (will break enpkg)")
    p.add_argument("--requirements", help=argparse.SUPPRESS)
    p.add_argument("--revert", metavar="REV#",
                   help="revert to a previous set of packages (does not revert "
                   "enstaller itself)")
    p.add_argument('-q', "--quiet", action="store_true",
                   help="Quiet output.")
    p.add_argument('-s', "--search", action="store_true",
                   help="search the online repo index "
                        "and display versions available")
    p.add_argument("--sys-config", action="store_true",
                   help="Do nothing, kept for backwarc compatibility.")
    p.add_argument("--sys-prefix", action="store_true",
                   help="use sys.prefix as the install prefix")
    p.add_argument("--update-all", action="store_true",
                   help="update all installed packages")
    p.add_argument("--user", action="store_true",
                   help="install into user prefix, i.e. --prefix=%r" %
                        _user_base())
    p.add_argument("--userpass", action="store_true",
                   help="prompt for Enthought authentication, and save in "
                   "configuration file .enstaller4rc")
    p.add_argument('-v', "--verbose", action="count", default=0,
                   help="Verbose output if specified once, more verbose if "
                        "specified twice, logs sent/received http headers if "
                        "specified thrice.")
    p.add_argument('--version', action="version",
                   version='enstaller version: ' + enstaller.__version__)
    p.add_argument("--whats-new", action="store_true",
                   help="display available updates for installed packages")
    p.add_argument("-y", "--yes", action="store_true",
                   help="Assume 'yes' to all queries and do not prompt.")
    p.add_argument("--no-autoupdate", action="store_true",
                   help="Do not attempt to autoupdate")

    return p


def _compute_reqs(cnames):
    reqs = []
    for arg in cnames:
        if '-' in arg:
            name, version = arg.split('-', 1)
            reqs.append(Requirement(name + ' ' + version))
        else:
            reqs.append(Requirement(arg))
    return reqs


def _enable_logging(verbosity_level):
    if verbosity_level >= 2:
        level = logging.DEBUG
    elif verbosity_level == 1:
        level = logging.INFO
    else:
        level = logging.WARN
    logging.basicConfig(level=level, format="%(message)s")
    if verbosity_level >= 3:
        http_client.HTTPConnection.debuglevel = 1

    msg = ("Using enstaller version {0!r} (git rev {1!r})".
           format(enstaller.__version__, enstaller.__git_revision__))
    logger.info(msg)


def _preprocess_options(argv):
    p = _create_parser()
    args = p.parse_args(argv)

    # Check for incompatible actions and options
    # Action options which take no package name pattern:
    simple_standalone_actions = (args.config, args.env, args.userpass,
                                 args.revert, args.log, args.whats_new,
                                 args.update_all, args.remove_enstaller,
                                 args.add_url, args.freeze, args.requirements)
    # Action options which can take a package name pattern:
    complex_standalone_actions = (args.list, args.imports,
                                  args.search, args.info, args.remove)

    count_simple_actions = sum(bool(opt) for opt in simple_standalone_actions)
    count_complex_actions = sum(bool(opt) for opt in complex_standalone_actions)

    if count_simple_actions + count_complex_actions > 1:
        p.error('Multiple action options specified')
    if count_simple_actions > 0 and len(args.cnames) > 0:
        p.error("Option takes no arguments")

    _enable_logging(args.verbose)

    if args.user:
        args.prefix = _user_base()

    if args.prefix and args.sys_prefix:
        p.error("Options --prefix and --sys-prefix exclude each other")

    if args.force and args.forceall:
        p.error("Options --force and --forceall exclude each other")

    return p, args


def _ensure_config_path():
    config_filename = _get_config_candidate()

    if config_filename is None:
        config_filename = HOME_ENSTALLER4RC
        write_default_config(config_filename)
    return config_filename


def _ensure_config_or_die():
    config_filename = _ensure_config_path()

    try:
        config = Configuration.from_file(config_filename)
    except InvalidConfiguration as e:
        print(str(e))
        sys.exit(EXIT_ABORTED)

    return config


def _compute_prefixes(args, config):
    sys_prefix = os.path.normpath(sys.prefix)

    # make prefix
    if args.sys_prefix:
        prefix = sys_prefix
    elif args.prefix:
        prefix = os.path.normpath(args.prefix)
    else:
        prefix = config.prefix

    # now make prefixes
    if prefix == sys_prefix:
        prefixes = [sys_prefix]
    else:
        prefixes = [prefix, sys_prefix]

    return prefix, prefixes


def main(argv=None):
    if argv is None:  # pragma: no cover
        argv = sys.argv[1:]

    parser, args = _preprocess_options(argv)

    pat = None
    if (args.list or args.search) and args.cnames:
        pat = re.compile(args.cnames[0], re.I)

    if args.config_path:
        config_filename = os.path.expanduser(args.config_path)
        try:
            config = Configuration.from_yaml_filename(config_filename)
        except IOError:
            msg = "Error: File {0!r} could not be read".format(config_filename)
            print(msg)
            sys.exit(-1)
        else:
            if config.auth is None:
                print("Authentication missing from {0!r}".format(config_filename))
                sys.exit(-1)
            use_new_format = True
    else:
        config = _ensure_config_or_die()
        config_filename = config.filename
        use_new_format = False

    setup_proxy_or_die(config, args.proxy)

    prefix, prefixes = _compute_prefixes(args, config)
    config.update(prefix=prefix)

    if args.user:
        try:
            check_prefixes(prefixes)
        except InvalidPythonPathConfiguration:
            msg = "Using the --user option, but your PYTHONPATH is not " \
                  "setup accordingly"
            warnings.warn(msg)

    exit_if_root_on_non_owned(args.yes)

    logger.info("prefixes")
    for prefix in prefixes:
        logger.info('    %s%s', prefix, ['', ' (sys)'][prefix == sys.prefix])

    if hasattr(args, "insecure"):
        config.update(verify_ssl=not args.insecure)

    if hasattr(args, "max_retries"):
        config.update(max_retries=args.max_retries)

    with Session.from_configuration(config) as session:
        if dispatch_commands_without_enpkg(args, config, config_filename,
                                           prefixes, prefix, pat,
                                           session):
            return

        if config.auth is None:
            configure_authentication_or_exit(config, config_filename,
                                             session)
        ensure_authenticated_config(config, config_filename, session,
                                    use_new_format=use_new_format)

        repository = repository_factory(session, config.indices, args.quiet)
        if args.quiet:
            progress_bar_context = None
        else:
            def fetch_progress_factory(*a, **kw):
                return console_progress_manager_factory(*a, show_speed=True,
                                                        **kw)

            progress_bar_context = ProgressBarContext(
                console_progress_manager_factory, fetch=fetch_progress_factory)
        enpkg = Enpkg(repository, session, prefixes, progress_bar_context,
                      args.force or args.forceall,
                      max_retries=config.max_retries)

        dispatch_commands_with_enpkg(args, enpkg, config, prefix, session, parser,
                                     pat)


def main_noexc(argv=None):
    # FIXME: re-enable traceback hiding (aka enstaller_debug=False) once
    # we have enstaller logs
    enstaller_debug = True

    try:
        main(argv)
        sys.exit(0)
    except KeyboardInterrupt:
        sys.exit(EXIT_ABORTED)
    except Exception as e:
        msg = """\
%s: Error: %s crashed (uncaught exception %s: %s).
Please report this on enstaller issue tracker:
    http://github.com/enthought/enstaller/issues"""
        if enstaller_debug:
            raise
        else:
            msg += "\nYou can get a full traceback by setting the " \
                   "ENSTALLER_DEBUG environment variable"
            print(msg % ("enstaller", "enstaller", e.__class__, repr(e)))
            sys.exit(1)

if __name__ == '__main__':  # pragma: no cover
    main_noexc()



import re

from enstaller.egg_meta import is_valid_eggname, split_eggname


class Requirement(object):
    """
    A requirement object is initialized by a requirement string. Attributes:
    name: the lowercase project name
    version: the list of possible versions required
    strictness: the level of strictness
        0   nothing matters, anything matches
        1   only the name must match
        2   name and version must match
        3   name, version and build must match
    """
    @classmethod
    def from_anything(cls, arg):
        if isinstance(arg, cls):
            return arg
        elif is_valid_eggname(arg):
            return cls('%s %s-%d' % split_eggname(arg))
        else:
            return cls(arg)

    pat = re.compile(r'(?:([\w.]+)(?:\s+([\w.]+)(?:-(\d+))?)?)?$')

    def __init__(self, req_string):
        m = self.pat.match(str(req_string.strip()))
        if m is None:
            raise Exception("Not a valid requirement: %r" % req_string)
        self.name, self.version, self.build = m.groups()
        self.strictness = 0
        if self.name is not None:
            self.name = self.name.lower()
            self.strictness = 1
        if self.version is not None:
            self.strictness = 2
        if self.build is not None:
            self.build = int(self.build)
            self.strictness = 3

    def as_dict(self):
        res = {}
        for var_name in 'name', 'version', 'build':
            if getattr(self, var_name):
                res[var_name] = getattr(self, var_name)
        return res

    def matches(self, candidate):
        """
        Returns whether the given package candidate matches the requirement.

        Parameters
        ----------
        candidate: PackageVersionInfo
            The package to test.
        """
        if self.strictness == 0:
            return True
        if candidate.name != self.name:
            return False
        if self.strictness == 1:
            return True
        if str(candidate.version.upstream) != self.version:
            return False
        if self.strictness == 2:
            return True
        return candidate.version.build == self.build

    def __str__(self):
        if self.strictness == 0:
            return ''
        res = self.name
        if self.version:
            res += ' %s' % self.version
        if self.build:
            res += '-%d' % self.build
        return res

    def __repr__(self):
        """
        return a canonical representation of the object
        """
        return 'Requirement(%r)' % str(self)

    def __eq__(self, other):
        return (self.name == other.name and
                self.version == other.version and
                self.build == other.build and
                self.strictness == other.strictness)

    def __hash__(self):
        return (hash(self.strictness) ^ hash(self.name) ^
                hash(self.version) ^ hash(self.build))
